import re
from datetime import date
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict, get_args
from ..util import validate_regexp
from ..constant import Gender


CHECKSUM_TYPE = Literal['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                        'A', 'B', 'C', 'D', 'E', 'F', 'H', 'J', 'K', 'L',
                        'M', 'N', 'P', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
                        'Y']
"""literal for checksum type"""


class ParseResult(TypedDict):
    """The parse result of Finland personal identity code, HETU"""
    yyyymmdd: date
    """Birthday"""
    gender: Gender
    """Gender, male or female"""
    sn: str
    """The serial number born at the same date"""
    checksum: CHECKSUM_TYPE
    """checksum code in str"""


class PersonalIdentityCode:
    """
    Finland personal identity code, HETU
    https://en.wikipedia.org/wiki/National_identification_number#Finland
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'FI',
        'min_length': 11,
        'max_length': 11,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<dd>\d{2})(?P<mm>\d{2})(?P<yy>\d{2})'
                             r'(?P<century>[-+ABCDEFUVWXY])'
                             r'(?P<sn>\d{3})'
                             r'(?P<check>[0-9A-Z])$'),
        'alias_of': None,
        'names': ['personal identity code',
                  'HETU'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Finland'],
        'deprecated': False
    })

    DOB_BASE_MAP = {
        '+': 1800,
        '-': 1900,
        'U': 1900,
        'V': 1900,
        'W': 1900,
        'X': 1900,
        'Y': 1900,
        'A': 2000,
        'B': 2000,
        'C': 2000,
        'D': 2000,
        'E': 2000,
        'F': 2000,
    }
    """ The century map for id """

    CHECKSUM_LIST = list(get_args(CHECKSUM_TYPE))
    """ possible checksum characteres """

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the FIN id number
        """
        if not validate_regexp(id_number, PersonalIdentityCode.METADATA.regexp):
            return False
        return PersonalIdentityCode.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """ parse the FIN HETU id"""
        match_obj = PersonalIdentityCode.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        elif not PersonalIdentityCode.checksum(id_number):
            return None
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        century = match_obj.group('century')
        sn = match_obj.group('sn')
        yyyy_base = PersonalIdentityCode.DOB_BASE_MAP[century]
        try:
            return {
                'yyyymmdd': date(yyyy_base + yy, mm, dd),
                'gender': Gender.MALE if int(sn) % 2 == 1 else Gender.FEMALE,
                'sn': sn,
                'checksum': match_obj.group('check')
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number: str) -> bool:
        """check if the ID valid against its checksum"""
        match_obj = PersonalIdentityCode.METADATA.regexp.match(id_number)
        if not match_obj:
            return False
        numbers = int(match_obj.group('dd') + match_obj.group('mm') + match_obj.group('yy') + match_obj.group('sn'))
        check_digit = PersonalIdentityCode.CHECKSUM_LIST[numbers % 31]
        return match_obj.group('check') == check_digit



from .nzl.driver_license import DriverLicenseNumber
from .nzl.inland_revenue_department import InlandRevenueDepartmentNumber
from .nzl.passport import PassportNumber
from .nzl.health_index import NationalHealthIndexNumber
from .util import alias_of

NationalID = alias_of(DriverLicenseNumber)



import re
from types import SimpleNamespace

from ..util import validate_regexp, weighted_modulus_digit


class MyNumber:
    """
    Japan my number format
    https://en.wikipedia.org/wiki/National_identification_number#Japan
    https://tin-check.com/en/
    https://github.com/kufu/tsubaki/blob/433d65aac341bcd58e7d8141f3f4ac374977617f/lib/tsubaki/my_number.rb#L12
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'JP',
        # length without insignificant chars
        'min_length': 12,
        'max_length': 12,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(\d{12}$)'),
        'alias_of': None,
        'names': ['National ID Number',
                  'My Number',
                  'マイナンバー'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Japan'],
        'deprecated': False
    })

    MULTIPLIER = [6, 5, 4, 3, 2, 7, 6, 5, 4, 3, 2]
    """multiplier for checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate JPN national id number
        """
        if not validate_regexp(id_number, MyNumber.METADATA.regexp):
            return False
        return MyNumber.checksum(id_number) == id_number[-1]

    @staticmethod
    def checksum(id_number: str) -> str:
        """Calculate Japan national id checksum"""
        arr = [int(i) for i in id_number[:11]]
        rem = weighted_modulus_digit(arr, MyNumber.MULTIPLIER, 11, True)
        return str(0 if rem <= 1 else (11 - rem))



import re
from types import SimpleNamespace
from typing import Optional
from ..constant import Citizenship
from .national_id import NationalID, ParseResult


class OldIDParseResult(ParseResult):
    """old format contains more info"""
    citizenship: Citizenship


class OldNationalID:
    """
    LKA old National ID number format which is phased out on 2016
    # https://en.wikipedia.org/wiki/National_identification_number#Sri_Lanka
    # https://drp.gov.lk/Templates/Artical%20-%20English%20new%20number.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'LK',
        'min_length': 10,
        'max_length': 10,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<year>\d{2})'
                             r'(?P<days>\d{3})'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)'
                             r'(?P<citizenship>[XxVv])$'),
        'alias_of': None,
        'names': ['National ID Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Sri_Lanka',
                  'https://drp.gov.lk/Templates/Artical%20-%20English%20new%20number.html'],
        'deprecated': True
    })

    @staticmethod
    def to_new(id_number: str) -> Optional[str]:
        """convert the old format to the new format"""
        match_obj = OldNationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        year = match_obj.group('year')
        days = match_obj.group('days')
        sn = match_obj.group('sn')
        checksum = match_obj.group('checksum')
        return f'19{year}{days}0{sn}{checksum}'

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the old format id numbers
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return OldNationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[OldIDParseResult]:
        """it converts to new format and parse the extra citizen value"""
        new_id_num = OldNationalID.to_new(id_number)
        if not new_id_num:
            return None
        result = NationalID.parse(new_id_num)
        if not result:
            return None
        citizenship = Citizenship.CITIZEN if id_number[-1].upper() == 'V' else Citizenship.RESIDENT
        return {
            **result,
            'citizenship': citizenship
        }

    @staticmethod
    def checksum(id_number) -> bool:
        """use new format to check the checksum"""
        new_id_num = OldNationalID.to_new(id_number)
        if not new_id_num:
            return False
        return NationalID.checksum(new_id_num)



import re
from types import SimpleNamespace
from ..util import validate_regexp
from .util import normalize


class DriverLicenseNumber:
    """
    Australia driver license number format
    https://learn.microsoft.com/en-us/microsoft-365/compliance/sit-defn-australia-drivers-license-number?view=o365-worldwide
    https://techdocs.broadcom.com/us/en/symantec-security-software/information-security/data-loss-prevention/15-8/about-data-loss-prevention-policies-v27576413-d327e9/library-of-system-data-identifiers-v95989112-d327e56315/australia-driver-s-license-number-v130004514-d327e56830.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'AU',
        # length without insignificant chars
        'min_length': 6,
        'max_length': 10,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': False,
        # regular expression to validate the id
        'regexp': re.compile(r'^('
                             r'\d{9}|\d{3} \d{3} \d{3}|'
                             r'\d{8}|\d{2} \d{3} \d{3}|'
                             r'[A-Za-z]\d{5}|'
                             r'\d{10}|\d{3}-\d{3}-\d{4}'
                             r')$'),
        'alias_of': None,
        'names': ['Driver Licence Number'],
        'links': [
            'https://learn.microsoft.com/en-us/microsoft-365/compliance/'
            'sit-defn-australia-drivers-license-number?view=o365-worldwide',
            'https://techdocs.broadcom.com/us/en/symantec-security-software/information-security/'
            'data-loss-prevention/15-8/about-data-loss-prevention-policies-v27576413-d327e9/'
            'library-of-system-data-identifiers-v95989112-d327e56315/'
            'australia-driver-s-license-number-v130004514-d327e56830.html'],
        'deprecated': False
    })

    BLACK_TRAILING_NUMBER = ['00000', '11111', '22222', '33333', '44444', '55555', '66666', '77777', '88888', '99999']
    """black list for some numbers"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the AUS driver license number
        """
        if not validate_regexp(id_number, DriverLicenseNumber.METADATA.regexp):
            return False
        return normalize(id_number)[-5:] not in DriverLicenseNumber.BLACK_TRAILING_NUMBER



import re
from types import SimpleNamespace
from ..util import validate_regexp
from .util import normalize


class RGNumber:
    """
    Brazil Registro Geral Number
    https://en.wikipedia.org/wiki/National_identification_number#Brazil
    https://en.wikipedia.org/wiki/Brazilian_identity_card
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'BR',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^(\d{2}\.\d{3}\.\d{3}-[\d|X])$'),
        'alias_of': None,
        'names': ['RG number',
                  'Registro Geral number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Brazil',
                  'https://en.wikipedia.org/wiki/Brazilian_identity_cards'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the BRA Registro Geral Number
        """
        if not validate_regexp(id_number, RGNumber.METADATA.regexp):
            return False
        return RGNumber.checksum(id_number)

    MULTIPLIER = [2, 3, 4, 5, 6, 7, 8, 9]

    @staticmethod
    def checksum(id_number: str) -> bool:
        """Validate RG number checksum"""
        normalized = normalize(id_number)
        number_list = [int(char) for char in list(normalized[:8])]
        # X is equal to 11 in check digit
        check_digit = 11 if normalized[8] == 'X' else int(normalized[8])
        total = sum([value * RGNumber.MULTIPLIER[index] for (index, value) in enumerate(number_list)])
        return True if ((total + check_digit * 100) % 11) == 0 else False



from .kor.resident_registration import ResidentRegistration
from .kor.old_registration_registration import OldResidentRegistration
from .kor.arc import ARC
from .util import alias_of

NationalID = alias_of(ResidentRegistration)



import re
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import CHECK_DIGIT, validate_regexp
from .util import EntityType, EntityDivision, checksum


class BINParseResult(TypedDict):
    """The parse result of BIN"""
    yy: int
    """registration year (2 digits)"""
    mm: int
    """registration  month"""
    entity_type: EntityType
    """entity type"""
    entity_division: EntityDivision
    """type of this entity division"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """check digit"""


class BusinessIDNumber:
    """
    Kazakhstan Business identification number, Бизнес-идентификационный номер
    https://korgan-zan.kz/en/obtaining-iin-and-bin-in-kazakhstan/
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'KZ',
        'min_length': 12,
        'max_length': 12,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yy>\d{2})(?P<mm>\d{2})'
                             r'(?P<type>[4-6])'
                             r'(?P<division>[0-3])'
                             r'(?P<sn>\d{5})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Business Identification Number',
                  'Бизнес-идентификационный номер'],
        'links': ['https://korgan-zan.kz/en/obtaining-iin-and-bin-in-kazakhstan/'],
        'deprecated': False
    })

    ENTITY_TYPE_MAP = {
        '4': EntityType.ResidentEntity,
        '5': EntityType.ResidentEntity,
        '6': EntityType.ResidentEntity
    }

    DIVISION_TYPE_MAP = {
        '0': EntityDivision.HeadUnit,
        '1': EntityDivision.Branch,
        '2': EntityDivision.Representative,
        '3': EntityDivision.Peasant
    }

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate"""
        if not validate_regexp(id_number, BusinessIDNumber.METADATA.regexp):
            return False
        return BusinessIDNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[BINParseResult]:
        """parse the result"""
        match_obj = BusinessIDNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        bin_checksum = BusinessIDNumber.checksum(id_number)
        if bin_checksum is None or bin_checksum != int(match_obj.group('checksum')):
            return None
        entity_type = match_obj.group('type')
        division = match_obj.group('division')
        if entity_type not in BusinessIDNumber.ENTITY_TYPE_MAP:
            return None
        if division not in BusinessIDNumber.DIVISION_TYPE_MAP:
            return None
        return {
            'yy': int(match_obj.group('yy')),
            'mm': int(match_obj.group('mm')),
            'entity_type': BusinessIDNumber.ENTITY_TYPE_MAP[entity_type],
            'entity_division': BusinessIDNumber.DIVISION_TYPE_MAP[division],
            'sn': match_obj.group('sn'),
            'checksum': int(match_obj.group('checksum')),
        }

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        if not validate_regexp(id_number, BusinessIDNumber.METADATA.regexp):
            return False
        return checksum(id_number)



import re
from types import SimpleNamespace
from ..util import validate_regexp


class NationalHealthIndexNumber:
    """
    New Zealand national health index(NHI) number format
    https://techdocs.broadcom.com/us/en/symantec-security-software/information-security/data-loss-prevention/15-8/about-data-loss-prevention-policies-v27576413-d327e9/library-of-system-data-identifiers-v95989112-d327e56315/new-zealand-national-health-index-number-v117807810-d327e90250/new-zealand-national-health-index-number-narrow-br-v117808786-d327e90350.html
    This is a python version of this one: https://gist.github.com/mcshaz/b41dc6bd4aa3104d54da677e2b4f6b45
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'NZ',
        # length without insignificant chars
        'min_length': 7,
        'max_length': 7,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        # no I and no O in alphabet
        'regexp': re.compile(r'^('
                             r'[A-HJ-NP-Z]{3}\d{4}|'
                             r'[A-HJ-NP-Z]{3}\d{2}[A-HJ-NP-Z]{2}|'
                             r')$'),
        'alias_of': None,
        'names': ['National Health Index Number',
                  'NHI'],
        'links': ['https://techdocs.broadcom.com/us/en/symantec-security-software/information-security/'
                  'data-loss-prevention/15-8/about-data-loss-prevention-policies-v27576413-d327e9/'
                  'library-of-system-data-identifiers-v95989112-d327e56315/'
                  'new-zealand-national-health-index-number-v117807810-d327e90250/'
                  'new-zealand-national-health-index-number-narrow-br-v117808786-d327e90350.html'],
        'deprecated': False
    })

    ALPHABET_LIST = list('ABCDEFGHJKLMNPQRSTUVWXYZ')

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the NZL NHI number
        """
        return NationalHealthIndexNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://gist.github.com/mcshaz/b41dc6bd4aa3104d54da677e2b4f6b45"""
        if not validate_regexp(id_number, NationalHealthIndexNumber.METADATA.regexp):
            return False
        check_digit = id_number[-1]
        source_list = list(id_number[:-1])
        total = 0
        for (index, char) in enumerate(source_list):
            if char in NationalHealthIndexNumber.ALPHABET_LIST:
                decimal = NationalHealthIndexNumber.ALPHABET_LIST.index(char) + 1
            else:
                decimal = int(char)
            total += decimal * (7 - index)
        if check_digit in NationalHealthIndexNumber.ALPHABET_LIST:
            # new NHI format
            modulus = total % 24
            return NationalHealthIndexNumber.ALPHABET_LIST[23 - modulus] == check_digit
        else:
            check_decimal = int(check_digit)
            # old NHI format
            modulus = total % 11
            if modulus == 0:
                return False
            return check_decimal == 0 if modulus == 1 else check_decimal == (11 - modulus)



from .che.social_security import SocialSecurityNumber
from .che.business_id import BusinessID
from .util import alias_of

AVH = alias_of(SocialSecurityNumber)
"""alias of SocialSecurityNumber"""
NationalID = alias_of(SocialSecurityNumber)
"""alias of SocialSecurityNumber"""

UID = alias_of(BusinessID)
"""alias of Business ID"""



import re
from types import SimpleNamespace
from ..util import validate_regexp
from .util import normalize


class CPFNumber:
    """
       Brazil CPF number
       https://en.wikipedia.org/wiki/National_identification_number#Brazil
       """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'BR',
        # length without insignificant chars
        'min_length': 11,
        'max_length': 11,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^(\d{3}\.?\d{3}\.?\d{3}-?\d{2})$'),
        'alias_of': None,
        'names': ['CPF number',
                  'Cadastro de Pessoas Físicas'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Brazil'],
        'deprecated': False
    })

    MULTIPLIER1 = [10, 9, 8, 7, 6, 5, 4, 3, 2]
    MULTIPLIER2 = [11, 10, 9, 8, 7, 6, 5, 4, 3]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the BRA Cadastro de Pessoas Físicas Number
        https://en.wikipedia.org/wiki/CPF_number
        https://4app.net/tools/validator/document/cpf_validator
        """
        if not validate_regexp(id_number, CPFNumber.METADATA.regexp):
            return False
        return CPFNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """Validate CPF number checksum digits"""
        normalized = normalize(id_number)
        number_list = [int(char) for char in list(normalized[:9])]
        return normalized[9] == CPFNumber.first_digit_checksum(number_list) and normalized[
            10] == CPFNumber.second_digit_checksum(number_list)

    @staticmethod
    def first_digit_checksum(number_list) -> str:
        """Get the first checksum digit"""

        total = sum([value * CPFNumber.MULTIPLIER1[index] for (index, value) in enumerate(number_list)])
        return str(CPFNumber.get_checksum(total))

    @staticmethod
    def second_digit_checksum(number_list) -> str:
        """Get the second checksum digit"""

        first_checksum = CPFNumber.first_digit_checksum(number_list)
        total = sum([value * CPFNumber.MULTIPLIER2[index] for (index, value) in enumerate(number_list)])
        # Using first digit of the checksum to get total
        total += int(first_checksum) * 2
        return str(CPFNumber.get_checksum(total))

    @staticmethod
    def get_checksum(total: int) -> int:
        """Map the total sum to checksum number"""
        remainder = total % 11
        return 0 if remainder < 2 else 11 - remainder



import re
from types import SimpleNamespace
from typing import Optional, TypedDict
from datetime import date

from ..constant import Gender
from ..util import validate_regexp


class ParseResult(TypedDict):
    """parse result of National ID"""
    gender: Gender
    """gender, possible value: male, female"""
    yyyymmdd: date
    """dob"""
    checksum: str
    """checksum, 2 digits"""


class NationalID:
    """
    Norway National ID number
    https://en.wikipedia.org/wiki/National_identification_number#Norway
    https://en.wikipedia.org/wiki/National_identity_number_(Norway)
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'NO',
        # length without insignificant chars
        'min_length': 11,
        'max_length': 11,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<dd>\d{2})'
                             r'(?P<mm>\d{2})'
                             r'(?P<yy>\d{2})'
                             r'(?P<individual_number>\d{3})'
                             r'(?P<checksum>\d{2})$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'fødselsnummer',
                  'birth number',
                  'riegádannummir'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Norway',
                  'https://en.wikipedia.org/wiki/National_identity_number_(Norway)'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the NOR id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        if not NationalID.parse(id_number):
            return False
        return NationalID.checksum(id_number)

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        individual_code = match_obj.group('individual_number')
        yy = match_obj.group('yy')
        mm = match_obj.group('mm')
        dd = match_obj.group('dd')

        birth_century = '20'
        individual_num = int(individual_code)
        if 0 <= individual_num < 500:
            birth_century = 19
        elif 500 <= individual_num < 750 and int(yy) >= 54:
            birth_century = 18
        elif 900 <= individual_num < 1000 and int(yy) >= 40:
            birth_century = 19

        return {
            "gender": Gender.FEMALE if int(individual_code[2]) % 2 == 0 else Gender.MALE,
            'yyyymmdd': date(int(f'{birth_century}{yy}'), int(mm), int(dd)),
            "checksum": match_obj.group('checksum')
        }

    FIRST_MAGIC_MULTIPLIER = [3, 7, 6, 1, 8, 9, 4, 5, 2, 1]
    SECOND_MAGIC_MULTIPLIER = [5, 4, 3, 2, 7, 6, 5, 4, 3, 2, 1]

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://en.wikipedia.org/wiki/National_identity_number_(Norway)#Check_digits"""
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        number_list = [int(char) for char in id_number]
        # Digit 10th
        first_total = sum([value * number_list[idx] for (idx, value) in enumerate(NationalID.FIRST_MAGIC_MULTIPLIER)])
        # Digit 11th
        second_total = sum([value * number_list[idx] for (idx, value) in enumerate(NationalID.SECOND_MAGIC_MULTIPLIER)])
        return first_total % 11 == 0 and second_total % 11 == 0



import re
from types import SimpleNamespace
from ..util import validate_regexp, weighted_modulus_digit, letter_to_number


def normalize(id_number: str) -> str:
    """strip out useless characters/whitespaces"""
    return id_number.replace('/', '')


class PersonalPublicServiceNumber:
    """
    Ireland Personal Public Service Number
    https://en.wikipedia.org/wiki/Personal_Public_Service_Number
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'IE',
        'min_length': 8,
        'max_length': 10,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^\d{7}[A-W][A-W\s]?$|'
                             r'^\d{7}[A-W]/[A-W\s]?$'),
        'alias_of': None,
        'names': ['Personal Public Service Number',
                  'PPS',
                  'Uimhir Phearsanta Seirbhíse Poiblí',
                  'Uimh. PSP',
                  'Revenue and Social Insurance Number',
                  'RSI No'],
        'links': ['https://en.wikipedia.org/wiki/Personal_Public_Service_Number'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER = [8, 7, 6, 5, 4, 3, 2, 9]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the IRL personal public service number
        """
        if not validate_regexp(id_number, PersonalPublicServiceNumber.METADATA.regexp):
            return False
        return PersonalPublicServiceNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://en.wikipedia.org/wiki/Personal_Public_Service_Number#Check_character"""
        normalized = normalize(id_number)
        number_list = [int(i) for i in normalized[:7]]
        if len(normalized) == 9 and normalized[-1] not in [' ', 'W']:
            number_list.append(letter_to_number(normalized[-1]))
        modulus = weighted_modulus_digit(numbers=number_list,
                                         weights=PersonalPublicServiceNumber.MAGIC_MULTIPLIER,
                                         divider=23, modulus_only=True)
        # the last digit is a check_char if the length is 8
        check_char = normalized[-2] if len(normalized) == 9 else normalized[-1]

        return modulus == letter_to_number(check_char) % 23



from .kaz.business_id import BusinessIDNumber
from .kaz.individual_id import IndividualIDNumber
from .kaz.util import EntityType, EntityDivision
from .util import alias_of

NationalID = alias_of(IndividualIDNumber)
"""alias of IndividualIDNumber"""



import re
from datetime import date
from typing import Optional, TypedDict
from types import SimpleNamespace

from ..constant import Gender
from ..util import CHECK_DIGIT, validate_regexp


class BirthNumberParseResult(TypedDict):
    yyyymmdd: date
    """birthday of this ID, there is no way to know the century of the birthday. So, yy < 50 is 20yy else 19yy."""
    gender: Gender
    """only male or female"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """checksum"""


class BirthNumber:
    """
    Slovakia Birth Number format, rodné číslo (RČ)
    https://en.wikipedia.org/wiki/National_identification_number#Slovakia
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'SK',
        'min_length': 10,
        'max_length': 10,
        # length without insignificant chars
        'parsable': True,
        # has parse function
        'checksum': True,
        # has checksum function
        'regexp': re.compile(r'^(?P<yy>\d{2})'
                             r'(?P<mm>\d{2})'
                             r'(?P<dd>\d{2})/?'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)$'),
        # regular expression to validate the id
        'alias_of': None,
        'names': ['Birth Number',
                  'rodné číslo',
                  'RČ'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Slovakia'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate birth number
        """
        if not validate_regexp(id_number, BirthNumber.METADATA.regexp):
            return False
        return BirthNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[BirthNumberParseResult]:
        """
        parse the id number
        """
        match_obj = BirthNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        if not BirthNumber.checksum(id_number):
            return None

        yy = int(match_obj.group('yy'))
        mm_code = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        mm = mm_code if mm_code < 50 else mm_code - 50
        """
        from https://en.wikipedia.org/wiki/National_identification_number#Czech_Republic_and_Slovakia
        In a law that took place in the year 2004, a failsafe system has been implemented, where in case
        all valid serial numbers get depleted for a day, the number 20 gets added to the value of XX.
        This means that XX can get up to 32 for males, and 82 for females.
        """
        mm = mm - 20 if mm > 20 else mm
        year_base = 2000 if yy < 50 else 1900
        try:
            return {
                'yyyymmdd': date(year_base + yy, mm, dd),
                'gender': Gender.MALE if mm_code < 50 else Gender.FEMALE,
                'sn': match_obj.group('sn'),
                'checksum': int(match_obj.group('checksum'))
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number: str) -> bool:
        """
        Calculate SVK BirthNumber checksum digit
        """
        if not validate_regexp(id_number, BirthNumber.METADATA.regexp):
            return False
        return int(BirthNumber.normalize(id_number)) % 11 == 0

    @staticmethod
    def normalize(id_number: str) -> str:
        """remove the / out"""
        return re.sub(r'/', '', id_number)



import re
from types import SimpleNamespace
from ..util import validate_regexp


class IdentityCard:
    """
    Greece Identity Card, the new one.
    https://en.wikipedia.org/wiki/National_identification_number#Greece
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'GR',
        'min_length': 7,
        'max_length': 7,
        'parsable': False,
        'checksum': False,
        'regexp': re.compile(r'^[ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩABEZHIKMNOPTYX]{2}-?\d{6}$'),
        # They are two different char set, the former is Greek alphabet, the latter is Latin alphabet
        'alias_of': None,
        'names': ['Identity Card Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Greece'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate with regexp
        """
        return validate_regexp(id_number, IdentityCard.METADATA.regexp)



import re
from types import SimpleNamespace
from typing import Optional

from ..util import CHECK_DIGIT, validate_regexp, luhn_digit


class NationalID:
    """
    Israel Identity Number, מספר זהות, Mispar Zehut
    https://en.wikipedia.org/wiki/National_identification_number#Israel
    https://taxid.pro/docs/countries/israel
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'IL',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(\d{9})$'),
        'alias_of': None,
        'names': ['Identity Number',
                  'מספר זהות',
                  'Mispar Zehut'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Israel',
                  'https://taxid.pro/docs/countries/israel'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return str(NationalID.checksum(id_number)) == id_number[-1]

    @staticmethod
    def checksum(id_number: str) -> Optional[CHECK_DIGIT]:
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return None
        """Calculate national id checksum"""
        numbers = [int(i) for i in id_number]
        return luhn_digit(numbers[:-1], False)



import re
from types import SimpleNamespace

from ..util import validate_regexp, mn_modulus_digit, modulus_overflow_mod10


class PersonalID:
    """
    Croatia Personal ID number format, OIB
    https://en.wikipedia.org/wiki/Personal_identification_number_(Croatia)
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'HR',
        'min_length': 11,
        'max_length': 11,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r"^\d{11}$"),
        'alias_of': None,
        'names': ['Personal ID Number',
                  'Osobni identifikacijski broj',
                  'OIB'],
        'links': ['https://en.wikipedia.org/wiki/Personal_identification_number_(Croatia)'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate HRV id number
        """
        if not validate_regexp(id_number, PersonalID.METADATA.regexp):
            return False
        return PersonalID.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """
        Calculate HRV personal id checksum digit
        """
        numbers = [int(char) for char in id_number]
        checksum = modulus_overflow_mod10(mn_modulus_digit(numbers[:-1], 10, 11))
        return numbers[-1] == checksum



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..constant import Citizenship
from ..util import validate_regexp


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'-', '', id_number)


class ParseResult(TypedDict):
    """parse result for NRIC"""
    yyyymmdd: date
    """birthday"""
    location: str
    """registration location"""
    citizenship: Citizenship
    """citizenship"""
    sn: str
    """serial number"""


class NRIC:
    """
    Malaysia National ID number format, NRIC
    https://en.wikipedia.org/wiki/Malaysian_identity_card#Structure_of_the_National_Registration_Identity_Card_Number_(NRIC)
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'MY',
        'min_length': 12,
        'max_length': 12,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})-?'
                             r'(?P<pb>\d{2})-?'
                             r'(?P<sn>\d{4})$'),
        'alias_of': None,
        'names': ['National Registration Identity Card Number',
                  'NRIC'],
        'links': ['https://en.wikipedia.org/wiki/Malaysian_identity_card#'
                  'Structure_of_the_National_Registration_Identity_Card_Number_(NRIC)'],
        'deprecated': False
    })

    WRONG_PB_CODE = ['00', '17', '18', '19', '20',
                     '69', '70', '73', '80', '81', '94', '95', '96', '97']
    """black list for pb code"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the id number
        """
        if not validate_regexp(id_number, NRIC.METADATA.regexp):
            return False
        return NRIC.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """pares the result"""
        match_obj = NRIC.METADATA.regexp.match(id_number)

        if not match_obj:
            return None
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        location = match_obj.group('pb')
        if not NRIC.check_location_code(location):
            return None
        sn = match_obj.group('sn')
        yyyy_base = 1900 if int(sn[0]) > 4 else 2000
        try:
            return {
                'yyyymmdd': date(yyyy_base + yy, mm, dd),
                'location': location,
                'citizenship': Citizenship.CITIZEN if int(location) < 60 else Citizenship.RESIDENT,
                'sn': sn
            }
        except ValueError:
            return None

    @staticmethod
    def check_location_code(location_code: str) -> bool:
        """we use blacklist to check wrong pb code"""
        return location_code not in NRIC.WRONG_PB_CODE



import re
from types import SimpleNamespace
from ..util import validate_regexp


class BusinessID:
    """
    Switzerland business identification number (UID)
    https://www.bfs.admin.ch/bfs/en/home/registers/enterprise-register/enterprise-identification/uid-general.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'CH',
        # length without insignificant chars
        'min_length': 12,
        'max_length': 12,
        'parsable': False,
        'checksum': False,
        'regexp': re.compile(r'^CHE-?\d{3}\.?\d{3}\.?\d{3}$'),
        'alias_of': None,
        'names': ['business identification number',
                  'UID'],
        'links': [
            'https://www.bfs.admin.ch/bfs/en/home/registers/enterprise-register/'
            'enterprise-identification/uid-general.html'
        ],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        return validate_regexp(id_number, BusinessID.METADATA.regexp)



import re
from types import SimpleNamespace

from ..util import validate_regexp


class SocialSecurityNumber:
    """
    United States Social Security number (SSN) format
    https://en.wikipedia.org/wiki/National_identification_number#United_States
    https://www.geeksforgeeks.org/how-to-validate-ssn-social-security-number-using-regular-expression/
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'US',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': False,
        # regular expression to validate the id
        'regexp': re.compile(r'^(?!666|000|9\d{2})\d{3}-(?!00)\d{2}-(?!0{4})\d{4}$'),
        'alias_of': None,
        'names': ['Social Security number',
                  'SSN'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#United_States'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate USA Social Security number
        """
        return validate_regexp(id_number, SocialSecurityNumber.METADATA.regexp)



import re
from types import SimpleNamespace
from ..util import validate_regexp, verhoeff_check


def normalize(id_number: str) -> str:
    """strip out useless characters/whitespaces"""
    return re.sub(r'[ -]', '', id_number)


class NationalID:
    """
    India National ID number format, UID
    https://en.wikipedia.org/wiki/National_identification_number#India
    https://archive.org/details/Aadhaar_numbering_scheme/page/n12/mode/1up?view=theater
    https://en.wikipedia.org/wiki/Verhoeff_algorithm#Table-based_algorithm
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'IN',
        'min_length': 12,
        'max_length': 12,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^[2-9]\d{3}[ -]?\d{4}[ -]?\d{4}$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'Unique Identification Number',
                  'UID'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#India',
                  'https://archive.org/details/Aadhaar_numbering_scheme/page/n12/mode/1up?view=theater'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the id number
        """
        return NationalID.checksum(id_number)

    @staticmethod
    def checksum(id_number) -> bool:
        """use verhoeff checksum"""
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return verhoeff_check([int(char) for char in normalize(id_number)])



import re
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..constant import Gender
from ..util import validate_regexp


class ParseResult(TypedDict):
    """The parse result of Pakistan ID"""
    location: str
    """registration location"""
    sn: str
    """serial number"""
    gender: Gender
    """gender: male or female"""


class NationalID:
    """
    Pakistan National ID Card number format, CNIC, NIC, قومی شناختی کارڈ
    https://en.wikipedia.org/wiki/National_identification_number#Pakistan
    https://en.wikipedia.org/wiki/CNIC_(Pakistan)#Security_features
    https://www.geo.tv/latest/250118-mystery-behind-13-digit-cnic-number
    https://www.informationpk.com/interesting-information-about-or-meaning-of-nadra-cnic-13-digits-number/
    check website: https://cnic.com.pk/
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'PA',
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<location>\d{5})-?'
                             r'(?P<sn>\d{7})-?'
                             r'(?P<gender>\d)$'),
        'alias_of': None,
        'names': ['National ID Card Number',
                  'CNIC',
                  'NIC',
                  'قومی شناختی کارڈ'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Pakistan',
                  'https://en.wikipedia.org/wiki/CNIC_(Pakistan)#Security_features',
                  'https://www.geo.tv/latest/250118-mystery-behind-13-digit-cnic-number',
                  'https://www.informationpk.com/'
                  'interesting-information-about-or-meaning-of-nadra-cnic-13-digits-number/'],
        'deprecated': False
    })
    """metadata of this id"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate"""
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        try:
            return {
                'location': match_obj.group('location'),
                'sn': match_obj.group('sn'),
                'gender': Gender.MALE if int(match_obj.group('gender')) % 2 == 1 else Gender.FEMALE,
            }
        except ValueError:
            # catch the date error
            return None


CNIC = NationalID
"""alias of NationalID"""



import re
from enum import Enum
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict
from ..util import weighted_modulus_digit, modulus_overflow_mod10, validate_regexp


class ThaiCitizenship(Enum):
    """Thailand's citizenship type"""
    OTHER = 0
    # Not used for Thai nationals, occasionally on other cards
    CITIZEN_AFTER_1984 = 1
    # Thai Nationals. Born after 1st Jan 1984
    CITIZEN_AFTER_1984_LATE_REGISTERED = 2
    # Thai Nationals. Born after 1st Jan 1984. Birth notified late
    CITIZEN_BEFORE_1984 = 3
    # Thai Nationals. Born & registered before 1st Jan 1984
    CITIZEN_BEFORE_1984_LATE_REGISTERED = 4
    # Thai Nationals. Born before 1st Jan 1984. Registered late.
    CITIZEN_SPECIAL_CASE = 5
    # Thai Nationals. Missed census or special cases.
    FOREIGN_RESIDENT = 6
    # Foreign Nationals living temporarily, or illegal migrants.
    FOREIGN_RESIDENT_CHILDREN = 7
    # Children of #6 who were born in Thailand.
    PERMANENT_RESIDENT = 8
    # Foreign Nationals living permanently, or Thai nationals by naturalisation.


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[ \-/]', '', id_number)


class ParseResult(TypedDict):
    """parse result of national id"""
    citizenship: ThaiCitizenship
    """Thailand specialized citizenship"""
    province_code: str
    """registration province"""
    district_code: str
    """registration district"""
    sn: str
    """serial number"""
    checksum: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """check digit"""


class NationalID:
    """
    THA National ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Thailand
    https://thailandformats.com/idcards

    This is the python version of https://github.com/awcode/thai-laravel
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'TH',
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<citizenship>[0-8])[ -]?'
                             r'(?P<province>\d{2})'
                             r'(?P<district>\d{2})[ -]?'
                             r'(?P<sn>\d{5}[ -]?\d{2})[ -]?'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'Population Identification Code',
                  'บัตรประชาชน',
                  'รหัสบัตรประชาชน'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Thailand',
                  'https://learn.microsoft.com/en-us/microsoft-365/compliance/'
                  'sit-defn-thai-population-identification-code?view=o365-worldwide'],
        'deprecated': False
    })

    PROVINCE_LIST = ['10', '11', '12', '13', '14', '15', '16', '17', '18', '19',
                     '20', '21', '22', '23', '24', '25', '26', '27',
                     '30', '31', '32', '33', '34', '35', '36', '37', '38', '39',
                     '40', '41', '42', '43', '44', '45', '46', '47', '48', '49',
                     '50', '51', '52', '53', '54', '55', '56', '57', '58',
                     '60', '61', '62', '63', '64', '65', '66', '67',
                     '70', '71', '72', '73', '74', '75', '76', '77',
                     '80', '81', '82', '83', '84', '85', '86',
                     '90', '91', '92', '93', '94', '95', '96']
    """possible province value"""
    DISTRICT_MAX_VALUE = {
        '10': 50, '11': 6, '12': 6, '13': 7, '14': 46, '15': 7, '16': 11, '17': 9, '18': 8, '19': 12,
        '20': 11, '21': 8, '22': 10, '23': 7, '24': 11, '25': 9, '26': 4, '27': 9,
        '30': 32, '31': 23, '32': 17, '33': 22, '34': 25, '35': 9, '36': 16, '37': 6, '38': 8, '39': 6,
        '40': 26, '41': 25, '42': 14, '43': 9, '44': 13, '45': 20, '46': 18, '47': 18, '48': 12, '49': 7,
        '50': 25, '51': 8, '52': 12, '53': 9, '54': 16, '55': 15, '56': 9, '57': 18, '58': 7,
        '60': 15, '61': 8, '62': 11, '63': 9, '64': 9, '65': 22, '66': 13, '67': 13,
        '70': 10, '71': 13, '72': 10, '73': 7, '74': 3, '75': 3, '76': 7, '77': 8,
        '80': 23, '81': 8, '82': 8, '83': 3, '84': 19, '85': 4, '86': 8,
        '90': 16, '91': 7, '92': 10, '93': 11, '94': 11, '95': 8, '96': 13,
    }
    """possible district max value"""
    DISTINCT_SPECIAL_CASE = {'44': [95]}
    """44 can have a special value"""

    MAGIC_MULTIPLIER = [13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2]
    """magic numbers for the checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the THA id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        citizenship = match_obj.group('citizenship')
        province = match_obj.group('province')
        district = match_obj.group('district')
        sn = normalize(match_obj.group('sn'))
        checksum = NationalID.checksum(id_number)
        if not NationalID.check_province_code(province):
            return None
        if not NationalID.check_district_code(province, district):
            return None
        if not checksum:
            return None
        else:
            return {
                'citizenship': ThaiCitizenship(int(citizenship)),
                'province_code': province,
                'district_code': district,
                'sn': sn,
                'checksum': int(match_obj.group('checksum'))
            }

    @staticmethod
    def checksum(id_number) -> bool:
        """algorithm: https://github.com/awcode/thai-laravel"""
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        # it uses modulus 11 algorithm with magic numbers
        numbers = [int(char) for char in normalize(id_number)]
        modulus = modulus_overflow_mod10(weighted_modulus_digit(numbers[:-1], NationalID.MAGIC_MULTIPLIER, 11))
        return modulus == numbers[-1]

    @staticmethod
    def check_province_code(province_code: str) -> bool:
        """check the province code"""
        return province_code in NationalID.PROVINCE_LIST

    @staticmethod
    def check_district_code(province_code: str, district_code: str) -> bool:
        """check the district code"""
        if district_code == '99':
            return True
        district_int = int(district_code)
        if district_int <= NationalID.DISTRICT_MAX_VALUE[province_code]:
            return True
        if province_code not in NationalID.DISTINCT_SPECIAL_CASE:
            return False
        return district_int in NationalID.DISTINCT_SPECIAL_CASE[province_code]



from copy import copy
from ..svk.birth_number import BirthNumber as SVKBirthNumber


CZE_METADATA = copy(SVKBirthNumber.METADATA)
CZE_METADATA.iso3166_alpha2 = 'CZ'
CZE_METADATA.links = ['https://en.wikipedia.org/wiki/National_identification_number#Czech_Republic_and_Slovakia']


class BirthNumber(SVKBirthNumber):
    """
    Czech Republic uses the same system of SVK. Birth Number (Czech/Slovak: rodné číslo (RČ))
    https://en.wikipedia.org/wiki/National_identification_number#Czech_Republic_and_Slovakia
    """
    METADATA = CZE_METADATA



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict, Tuple
from ..constant import Citizenship, Gender
from ..util import CHECK_DIGIT, weighted_modulus_digit, validate_regexp


class ParseResult(TypedDict):
    """parse result of CNP"""
    yyyymmdd: date
    """dob"""
    location: str
    """registration location"""
    gender: Gender
    """gender: male, female"""
    citizenship: Citizenship
    """citizenship or resident"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """checksum digit"""


class PersonalNumericalCode:
    """
    Personal Numerical Code (Cod Numeric Personal, CNP)
    https://en.wikipedia.org/wiki/National_identification_number#Romania
    https://en.wikipedia.org/wiki/Romanian_identity_card
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'RO',
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<gender_century>\d)'
                             r'(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<location>\d{2})'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Personal Numerical Code',
                  'Cod Numeric Personal',
                  'CNP',
                  'Carte de identitate'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Romania',
                  'https://en.wikipedia.org/wiki/Romanian_identity_card'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER = [2, 7, 9, 1, 4, 6, 3, 5, 8, 2, 7, 9]
    """multiplier for checksum"""

    YEAR_BASE_MAP = [1900, 1900, 1800, 1800, 2000, 2000]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the personal numerical code
        """
        if not validate_regexp(id_number, PersonalNumericalCode.METADATA.regexp):
            return False
        return PersonalNumericalCode.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the value"""
        match_obj = PersonalNumericalCode.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = PersonalNumericalCode.checksum(id_number)
        if not checksum:
            return None
        location = match_obj.group('location')
        if not (1 <= int(location) <= 52 or location == '99'):
            return None
        gender_century = int(match_obj.group('gender_century'))
        yy = int(match_obj.group('yy'))
        data = PersonalNumericalCode.get_gender_citizenship_year_base(gender_century, yy)
        if not data:
            return None
        gender, citizenship, year_base = data
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        sn = match_obj.group('sn')
        try:
            return {
                'yyyymmdd': date(year_base + yy, mm, dd),
                'location': location,
                'gender': gender,
                'citizenship': citizenship,
                'sn': sn,
                'checksum': int(match_obj.group('checksum'))
            }
        except ValueError:
            return None

    @staticmethod
    def get_gender_citizenship_year_base(gender_century: int, yy: int) -> Optional[Tuple[Gender, Citizenship, int]]:
        if gender_century > 8 or gender_century < 1:
            return None
        gender = Gender.MALE if gender_century % 2 == 1 else Gender.FEMALE
        citizenship = Citizenship.CITIZEN if gender_century < 7 else Citizenship.RESIDENT
        if gender_century < 7:
            year_base = PersonalNumericalCode.YEAR_BASE_MAP[gender_century]
        else:
            year_base = 2000 if yy < 50 else 1900
        return gender, citizenship, year_base

    @staticmethod
    def checksum(id_number) -> bool:
        """
        algorithm:
        https://en.wikipedia.org/wiki/National_identification_number#Romania
        """
        if not validate_regexp(id_number, PersonalNumericalCode.METADATA.regexp):
            return False
        numbers = [int(char) for char in id_number]
        modulus = weighted_modulus_digit(numbers[:-1], PersonalNumericalCode.MAGIC_MULTIPLIER, 11, True)
        if modulus == 10:
            modulus = 1
        return modulus == numbers[-1]



import re
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import CHECK_DIGIT, luhn_digit, validate_regexp


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[ \-/]', '', id_number)


class ParseResult(TypedDict):
    """Parse result of EmiratesID"""
    yyyy: int
    """year of birth"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """check digits"""


class EmiratesIDNumber:
    """
    ARE Emirates/Resident ID number format
    https://en.wikipedia.org/wiki/National_identification_number#United_Arab_Emirates
    This is the python version of https://gist.github.com/geordee/e51d111426de675c0c0f8503c2003047
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'AE',
        'min_length': 15,
        'max_length': 15,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^784[ -]?'
                             r'(?P<yyyy>\d{4})[ -]?'
                             r'(?P<sn>\d{7})[ -]?'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Emirates ID',
                  'Resident ID',
                  'رقم الهوية'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#United_Arab_Emirates'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the ARE id number
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return EmiratesIDNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = EmiratesIDNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        yyyy = match_obj.group('yyyy')
        sn = match_obj.group('sn')
        checksum_str = match_obj.group('checksum')
        checksum = EmiratesIDNumber.checksum(id_number)
        if checksum is None or str(checksum) != checksum_str:
            return None
        else:
            return {
                'yyyy': int(yyyy),
                'sn': sn,
                'checksum': checksum
            }

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        """use luhn algorithm to calculate the check digit"""
        if not validate_regexp(id_number, EmiratesIDNumber.METADATA.regexp):
            return None
        normalized = normalize(id_number)
        return luhn_digit([int(char) for char in normalized[:-1]])



import re
from types import SimpleNamespace
from idnumbers.nationalid.util import CHECK_DIGIT, validate_regexp, weighted_modulus_digit


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[-. ]', '', id_number)


def colombia_checksum(id_number: str) -> CHECK_DIGIT:
    """
    Python version of
    https://github.com/anghelvalentin/CountryValidator/blob/master/CountryValidator/CountriesValidators/ColombiaValidator.cs
    """
    numbers = [int(char) for char in list(normalize(id_number[:-1]))]
    numbers.reverse()
    modulus = weighted_modulus_digit(numbers, UniquePersonalID.WEIGHTS[0:len(numbers)], 11)
    if modulus == 11:
        return 0
    elif modulus == 10:
        return 1
    else:
        return modulus


class UniquePersonalID:
    """
    UniquePersonalID, NUIP, Número único de identidad personal
    https://en.wikipedia.org/wiki/Colombian_identity_card
    https://validatetin.com/colombia/#
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'CO',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 10,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(\d{2,3}\.?\d{3}\.?\d{3}-?\d)$'),
        'alias_of': None,
        'names': ['Unique Personal ID',
                  'NUIP',
                  'Número único de identidad personal'],
        'links': ['https://en.wikipedia.org/wiki/Colombian_identity_card',
                  'https://en.wikipedia.org/wiki/National_identification_number#Colombia',
                  'https://validatetin.com/colombia/#'],
        'deprecated': False
    })

    WEIGHTS = [3, 7, 13, 17, 19, 23, 29, 37, 41, 43, 47, 53, 59, 67, 71]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate id number
        """
        if not validate_regexp(id_number, UniquePersonalID.METADATA.regexp):
            return False
        return UniquePersonalID.checksum(id_number) == int(id_number[-1])

    @staticmethod
    def checksum(id_number: str) -> CHECK_DIGIT:
        if not validate_regexp(id_number, UniquePersonalID.METADATA.regexp):
            return False
        return colombia_checksum(id_number)



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, Tuple, TypedDict
from ..constant import Gender
from ..util import CHECK_DIGIT, validate_regexp
from .util import checksum


class IINParseResult(TypedDict):
    """The parse result of IIN"""
    yyyymmdd: date
    """dob"""
    gender: Gender
    """gender: male and female"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """check digit"""


class IndividualIDNumber:
    """
    Kazakhstan individual identification number, ЖСН, ZhSN, ИИН, IIN
    https://en.wikipedia.org/wiki/National_identification_number#Kazakhstan
    https://korgan-zan.kz/en/obtaining-iin-and-bin-in-kazakhstan/
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Kazakhstan-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'KZ',
        'min_length': 12,
        'max_length': 12,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<century>\d)'
                             r'(?P<sn>\d{4})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Individual Identification Number',
                  'ЖСН',
                  'ZhSN',
                  'ИИН',
                  'IIN'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Kazakhstan',
                  'https://korgan-zan.kz/en/obtaining-iin-and-bin-in-kazakhstan/',
                  'https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/Kazakhstan-TIN.pdf'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate"""
        if not validate_regexp(id_number, IndividualIDNumber.METADATA.regexp):
            return False
        return IndividualIDNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[IINParseResult]:
        """parse the result"""
        match_obj = IndividualIDNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        iin_checksum = IndividualIDNumber.checksum(id_number)
        if iin_checksum is None or iin_checksum != int(match_obj.group('checksum')):
            return None
        gender_year_base = IndividualIDNumber.get_gender_year_base(int(match_obj.group('century')))
        if not gender_year_base:
            return None
        gender, year_base = gender_year_base
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        try:
            return {
                'yyyymmdd': date(yy + year_base, mm, dd),
                'gender': gender,
                'sn': match_obj.group('sn'),
                'checksum': int(match_obj.group('checksum')),
            }
        except ValueError:
            # catch the date error
            return None

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        if not validate_regexp(id_number, IndividualIDNumber.METADATA.regexp):
            return False
        return checksum(id_number)

    @staticmethod
    def get_gender_year_base(century: CHECK_DIGIT) -> Optional[Tuple[Gender, int]]:
        gender = Gender.MALE if century % 2 == 1 else Gender.FEMALE
        if 1 <= century < 3:
            year_base = 1800
        elif 3 <= century < 5:
            year_base = 1900
        elif 5 <= century < 7:
            year_base = 2000
        else:
            return None
        return gender, year_base



from .aus.driver_license import DriverLicenseNumber
from .aus.medicare import MedicareNumber
from .aus.tax_file import TaxFileNumber
from .util import alias_of


NationalID = alias_of(DriverLicenseNumber)
"""use driver license as national id"""



import re
from types import SimpleNamespace
from ..util import validate_regexp
from .util import BLACK_TRAILING_NUMBER


class DriverLicenseNumber:
    """
    New Zealand driver license number format
    https://techdocs.broadcom.com/us/en/symantec-security-software/information-security/data-loss-prevention/15-8/about-data-loss-prevention-policies-v27576413-d327e9/library-of-system-data-identifiers-v95989112-d327e56315/new-zealand-driver-s-licence-number-v130004625-d327e90104/new-zealand-driver-s-licence-number-narrow-breadth-v130007408-d327e90179.html#v130007408
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'NZ',
        # length without insignificant chars
        'min_length': 8,
        'max_length': 8,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': False,
        # regular expression to validate the id
        'regexp': re.compile(r'^\w{2}\d{6}$'),
        'alias_of': None,
        'names': ['Driver License'],
        'links': ['https://techdocs.broadcom.com/us/en/symantec-security-software/information-security/'
                  'data-loss-prevention/15-8/about-data-loss-prevention-policies-v27576413-d327e9/'
                  'library-of-system-data-identifiers-v95989112-d327e56315/'
                  'new-zealand-driver-s-licence-number-v130004625-d327e90104/'
                  'new-zealand-driver-s-licence-number-narrow-breadth-v130007408-d327e90179.html#v130007408'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the NZL driver license number
        """
        if not validate_regexp(id_number, DriverLicenseNumber.METADATA.regexp):
            return False
        return id_number[-6:] not in BLACK_TRAILING_NUMBER



from copy import copy
from re import Pattern
from typing import List, Literal, Optional, Type, cast

VERHOEFF = {
    'D_TABLE': [
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        [1, 2, 3, 4, 0, 6, 7, 8, 9, 5],
        [2, 3, 4, 0, 1, 7, 8, 9, 5, 6],
        [3, 4, 0, 1, 2, 8, 9, 5, 6, 7],
        [4, 0, 1, 2, 3, 9, 5, 6, 7, 8],
        [5, 9, 8, 7, 6, 0, 4, 3, 2, 1],
        [6, 5, 9, 8, 7, 1, 0, 4, 3, 2],
        [7, 6, 5, 9, 8, 2, 1, 0, 4, 3],
        [8, 7, 6, 5, 9, 3, 2, 1, 0, 4],
        [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    ],

    'P_TABLE': [
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        [1, 5, 7, 6, 2, 8, 3, 0, 9, 4],
        [5, 8, 0, 3, 7, 9, 6, 1, 4, 2],
        [8, 9, 1, 6, 0, 4, 3, 5, 2, 7],
        [9, 4, 5, 3, 1, 2, 6, 8, 7, 0],
        [4, 2, 8, 6, 5, 7, 3, 9, 0, 1],
        [2, 7, 9, 3, 8, 0, 6, 4, 1, 5],
        [7, 0, 4, 6, 9, 1, 3, 2, 5, 8]
    ]
}
"""[Table](https://en.wikipedia.org/wiki/Verhoeff_algorithm#Table-based_algorithm) for the Verhoeff algorithm"""

CHECK_DIGIT: Type[int] = Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
"""Check digit type. Numeric check digits are only allowed in 0 to 9"""

CHECK_ALPHA = Literal['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K',
                      'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'Y', 'U',
                      'V', 'W', 'X', 'Y', 'Z']
"""Check digit type. Numeric check digits are only allowed in A to Z (all in upper cases)"""


def validate_regexp(id_number: str, regexp: Pattern[str]) -> bool:
    """validate string again the regular expression"""
    assert isinstance(id_number, str), 'id_number MUST be str'
    return regexp.search(id_number) is not None


def luhn_digit(digits: List[int], multipliers_start_by_two: bool = False) -> CHECK_DIGIT:
    """
    implement the algorithm of Luhn.
    https://en.wikipedia.org/wiki/Luhn_algorithm
    :param multipliers_start_by_two: Multipliers start by two
    :param digits: digits for calculating the check digit
    :return: checksum
    """
    total_sum = 0
    for idx, int_val in enumerate([0, *digits] if multipliers_start_by_two else digits):
        if idx % 2 == 0:
            total_sum += int_val
        elif int_val > 4:
            total_sum += (2 * int_val - 9)
        else:
            total_sum += (2 * int_val)
    return cast(CHECK_DIGIT, (10 - total_sum % 10) % 10)


def verhoeff_check(digits: List[int]) -> bool:
    """
    implement the verhoeff algorithm in table format:
    https://en.wikipedia.org/wiki/Verhoeff_algorithm#Table-based_algorithm
    """
    rev_digits = list(digits)
    rev_digits.reverse()
    c = 0
    for idx, num in enumerate(rev_digits):
        p_val = VERHOEFF["P_TABLE"][idx % 8][num]
        c = VERHOEFF["D_TABLE"][c][p_val]
    return c == 0


def weighted_modulus_digit(numbers: List[int], weights: Optional[List[int]], divider: int,
                           modulus_only: bool = False) -> int:
    """
    It metrix-multiples numbers and weights and calculate the modulus by the divider.
    :param numbers: the numbers list.
    :param weights: the weights list which will used in matrix multiplications. If weights is none, we use the
    [1] * len(numbers) as the weights.
    :param divider: the divider used for calculating modulus.
    :param modulus_only: If True, it returns the modulus calculated by divider, otherwise it returns divider - modulus.
    The default is False.
    :return: the value
    """
    if weights is None:
        weights = [1] * len(numbers)
    assert len(numbers) <= len(weights), 'numbers length must be less than or equal to weights length'
    modulus = sum([value * weights[index] for (index, value) in enumerate(numbers)]) % divider
    return modulus if modulus_only else divider - modulus


def mn_modulus_digit(numbers: List[int], m: int, n: int) -> int:
    """
    MN modulus check, (official name TBD) ISO 7064 mod 11 (n), 10 (m)?
    1. (adds numbers and product) mod by m
    2. next product = (2 * total) mod by n
    3. return n - product
    :param numbers: numbers
    :param m: M value used by calculate the first step
    :param n: N value used by the 2nd and 3rd step
    :return: the digit
    """
    product = m
    for number in numbers:
        total = (number + product) % m
        if total == 0:
            total = m
        product = (2 * total) % n

    return n - product


def modulus_overflow_mod10(modulus: int) -> CHECK_DIGIT:
    """
    get the units digit of a modulus. Some modulus may not be calculated with 10. In some cases, we need the units digit
    to be the ID.
    """
    return cast(CHECK_DIGIT, modulus % 10 if modulus > 9 else modulus)


def letter_to_number(letter: str, capital: bool = True):
    """
    English letter to its index. A = 1, B = 2...
    """
    assert len(letter) == 1 and letter.isalpha(), 'only allow one alphabet'
    if capital:
        return ord(letter) - 64
    return ord(letter) - 96


def ean13_digit(numbers: List[int]) -> CHECK_DIGIT:
    """
    The EAN-13 validation. The EAN-13 is a [barcode format](https://boxshot.com/barcode/tutorials/ean-13-barcodes/).
    This is the check digit of EAN-13.
    https://boxshot.com/barcode/tutorials/ean-13-calculator/
    """
    odd = 0
    even = 0
    for index, value in enumerate(numbers):
        if (index + 1) % 2 == 0:
            even += value
        else:
            odd += value
    total = even * 2 + odd
    modulus = total % 10
    return cast(CHECK_DIGIT,
                0 if modulus == 0 else (10 - modulus))


def alias_of(cls: Type) -> Type:
    assert hasattr(cls, 'METADATA'), f'the type {cls} must have METADATA attribute'
    metadata = copy(cls.METADATA)
    metadata.alias_of = cls

    class AliasType(cls):
        METADATA = metadata

    return AliasType



import re
from datetime import date
from typing import Optional, Union, TypedDict
from types import SimpleNamespace

from ..util import CHECK_DIGIT, validate_regexp
from .personal_code import PersonalCode


class OldParseResult(TypedDict):
    """parse result for the national id"""
    yyyymmdd: date
    """birthday"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """check digits"""


class OldPersonalCode:
    """
    Old Latvia personal code format, personas kods
    https://en.wikipedia.org/wiki/National_identification_number#Latvia
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Latvia-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'LV',
        # length without insignificant chars
        'min_length': 11,
        'max_length': 11,
        # has parse function
        'parsable': True,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(?P<dd>\d{2})(?P<mm>\d{2})(?P<yy>\d{2})-?'
                             r'(?P<century>\d)'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Personal Code',
                  'personas kods'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Latvia',
                  'https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/Latvia-TIN.pdf'],
        'deprecated': True
    })

    MULTIPLIER = [1, 6, 3, 7, 9, 10, 5, 8, 4, 2]
    """multiplier for checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate national id number
        """
        if not validate_regexp(id_number, OldPersonalCode.METADATA.regexp):
            return False
        return OldPersonalCode.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[OldParseResult]:
        """parse the result"""
        match_obj = OldPersonalCode.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = OldPersonalCode.checksum(id_number)
        if id_number[-1] != str(checksum):
            return None

        if match_obj.group('century') == '0':
            year_base = 1800
        elif match_obj.group('century') == '1':
            year_base = 1900
        elif match_obj.group('century') == '2':
            year_base = 2000
        else:
            return None

        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        sn = match_obj.group('sn')
        try:
            return {
                'yyyymmdd': date(yy + year_base, mm, dd),
                'sn': sn,
                'checksum': int(match_obj.group('checksum'))
            }
        except ValueError:
            # check of wrong date data
            return None

    @staticmethod
    def checksum(id_number: str) -> Optional[CHECK_DIGIT]:
        """Use new personal code to calculate the checksum"""
        return PersonalCode.checksum(id_number)



import re
from types import SimpleNamespace
from ..util import validate_regexp


def normalize(id_number: str) -> str:
    """strip out useless characters/whitespaces"""
    return re.sub(r'\.', '', id_number)


class BSN:
    """
    Netherlands National ID number
    Burgerservicenummer (BSN) (Citizen Service Number)
    https://en.wikipedia.org/wiki/National_identification_number#Netherlands
    https://nl.wikipedia.org/wiki/Burgerservicenummer
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'NL',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'(?!0000.00.000)^\d{4}\.\d{2}\.\d{3}$'),
        'alias_of': None,
        'names': ['Burgerservicenummer',
                  'BSN',
                  'Citizen Service Number',
                  'Personal Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Netherlands',
                  'https://nl.wikipedia.org/wiki/Burgerservicenummer'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the BSN id number
        """
        if not validate_regexp(id_number, BSN.METADATA.regexp):
            return False
        return BSN.checksum(id_number)

    MAGIC_MULTIPLIER = [9, 8, 7, 6, 5, 4, 3, 2]

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://nl.wikipedia.org/wiki/Burgerservicenummer#11-proef"""
        normalized = normalize(id_number)
        number_list = [int(char) for char in list(normalized[:-1])]
        total = sum([value * BSN.MAGIC_MULTIPLIER[index] for (index, value) in enumerate(number_list)])
        checksum = total % 11
        return str(total % 11) == normalized[-1] if checksum != 10 else False



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import CHECK_DIGIT, validate_regexp, weighted_modulus_digit


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[ -]', '', id_number)


class ParseResult(TypedDict):
    """The parse result of Iceland ID"""
    yyyymmdd: date
    """dob"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """check digit"""


class IcelandicID:
    """
    Iceland Icelandic identification number, kennitala
    https://en.wikipedia.org/wiki/Icelandic_identification_number
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'IS',
        'min_length': 10,
        'max_length': 10,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<dd>\d{2})(?P<mm>\d{2})(?P<yy>\d{2})-?'
                             r'(?P<sn>\d{2})'
                             r'(?P<checksum>\d)'
                             r'(?P<century>\d)$'),
        'alias_of': None,
        'names': ['Icelandic identification number',
                  'kennitala',
                  'kt.'],
        'links': ['https://en.wikipedia.org/wiki/Icelandic_identification_number'],
        'deprecated': False
    })

    WEIGHTS = [3, 2, 7, 6, 5, 4, 3, 2]

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate"""
        if not validate_regexp(id_number, IcelandicID.METADATA.regexp):
            return False
        return IcelandicID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = IcelandicID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = IcelandicID.checksum(id_number)
        if not checksum:
            return None
        if match_obj.group('century') == '9':
            year_base = 1900
        elif match_obj.group('century') == '0':
            year_base = 2000
        else:
            return None
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        try:
            return {
                'yyyymmdd': date(yy + year_base, mm, dd),
                'sn': match_obj.group('sn'),
                'checksum': int(match_obj.group('checksum')),
            }
        except ValueError:
            # catch the date error
            return None

    @staticmethod
    def checksum(id_number) -> bool:
        """
        check the checksum
        https://en.wikipedia.org/wiki/Icelandic_identification_number
        ref: https://github.com/aldavigdis/kennitala-gem/blob/main/lib/kennitala.rb#L295 for the 10
        """
        if not validate_regexp(id_number, IcelandicID.METADATA.regexp):
            return False
        numbers = [int(char) for char in normalize(id_number)]
        modulus = weighted_modulus_digit(numbers[0:-2], IcelandicID.WEIGHTS, 11, True)
        if modulus == 10:
            # ref https://github.com/aldavigdis/kennitala-gem/blob/main/lib/kennitala.rb#L295
            return False
        modulus = 0 if modulus == 0 else 11 - modulus
        return numbers[-2] == modulus



import re
from types import SimpleNamespace

from ..util import validate_regexp


class NationalID:
    """
    Papua New Guinea national id, NID
    https://en.wikipedia.org/wiki/National_identification_number#Papua_New_Guinea
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'PG',
        # length without insignificant chars
        'min_length': 10,
        'max_length': 10,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': False,
        # regular expression to validate the id
        'regexp': re.compile(r'^\d{10}$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'NID'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Papua_New_Guinea'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the NID
        """
        return validate_regexp(id_number, NationalID.METADATA.regexp)



import re
from types import SimpleNamespace

from ..util import validate_regexp, weighted_modulus_digit


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[ \-.]', '', id_number)


class FiscalInformationNumber:
    """
    Venezuela RIF (Fiscal Information Registry) number, Registro de Informacion Fiscal
    https://en.wikipedia.org/wiki/National_identification_number#Venezuela
    python version of
    https://github.com/anghelvalentin/CountryValidator/blob/master/CountryValidator/CountriesValidators/VenezuelaAfricaValidator.cs
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'VE',
        # length without insignificant chars
        'min_length': 10,
        'max_length': 10,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^[VEJPG]-?\d{8}-?\d$'),
        'alias_of': None,
        'names': ['Fiscal Information Number',
                  'RIF',
                  'Registro de Informacion Fiscal'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Venezuela'],
        'deprecated': False
    })

    WEIGHTS = [1, 3, 2, 7, 6, 5, 4, 3, 2]
    TYPE_MAP = {
        'V': 4,
        'E': 8,
        'J': 12,
        'P': 16,
        'G': 20
    }

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return FiscalInformationNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://github.com/therezor/ua-tax-number/blob/main/src/Decoder.php"""
        if not validate_regexp(id_number, FiscalInformationNumber.METADATA.regexp):
            return False
        normalized = normalize(id_number)
        numbers = [FiscalInformationNumber.TYPE_MAP[normalized[0]]] + [int(char) for char in list(normalized[1:])]
        modulus = weighted_modulus_digit(numbers[:-1], FiscalInformationNumber.WEIGHTS, 11)
        modulus = modulus if modulus < 10 else 0
        return modulus == numbers[-1]


RIF = FiscalInformationNumber
"""alias of FiscalInformationNumber"""



import re
from types import SimpleNamespace
from ..util import validate_regexp


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[-/]', '', id_number)


class TaxIDNumber:
    """
    Austria tax id number format
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Austria-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'AT',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(\d{2}-?\d{3}/?\d{4})$'),
        'alias_of': None,
        'names': ['Tax ID number'],
        'links': ['https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/Austria-TIN.pdf'],
        'deprecated': False
    })

    MULTIPLIER = [1, 2, 1, 2, 1, 2, 1, 2]
    """The multiplier for checksum"""
    OVERFLOW_SUM = {
        10: 1,
        12: 3,
        14: 5,
        16: 7,
        18: 9
    }
    """checksum algorithm shows we have to sum the two digits if the weighted value > 9"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the tax id number
        """
        if not validate_regexp(id_number, TaxIDNumber.METADATA.regexp):
            return False
        return TaxIDNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        if not validate_regexp(id_number, TaxIDNumber.METADATA.regexp):
            return False
        """algorithm: https://en.wikipedia.org/wiki/Tax_file_number#Check_digit"""
        normalized = normalize(id_number)
        numbers = [int(char) for char in list(normalized)]
        total = 0
        for (index, value) in enumerate(numbers[:-1]):
            weighted = value * TaxIDNumber.MULTIPLIER[index]
            if weighted in TaxIDNumber.OVERFLOW_SUM:
                weighted = TaxIDNumber.OVERFLOW_SUM[weighted]
            total += weighted
        checksum = (100 - total) % 10
        return checksum == numbers[-1]





from enum import Enum
from typing import Optional
from ..util import CHECK_DIGIT, weighted_modulus_digit


class EntityType(Enum):
    """Kazakhstan BIN entity type"""
    ResidentEntity = 'resident_entity'
    NonResidentEntity = 'non_resident_entity'
    IP = 'ip'


class EntityDivision(Enum):
    """Kazakhstan BIN entity division type"""
    HeadUnit = 'head_unit'
    Branch = 'branch'
    Representative = 'representative_office'
    Peasant = 'peasant'


WEIGHTS1 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
WEIGHTS2 = [3, 4, 5, 6, 7, 8, 9, 10, 11, 1, 2]


def checksum(id_number) -> Optional[CHECK_DIGIT]:
    """
    check the checksum
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Kazakhstan-TIN.pdf
    """
    numbers = [int(char) for char in id_number]
    modulus = weighted_modulus_digit(numbers[0:-1], WEIGHTS1, 11, True)
    if modulus == 10:
        modulus = weighted_modulus_digit(numbers[0:-1], WEIGHTS2, 11, True)
        # the second modulus will not be 10. If it is, it's wrong id number
    return modulus if modulus < 10 else None



import re
from types import SimpleNamespace
from ..util import validate_regexp


class PhilID:
    """
    Philippines PhilID Card Number, PCN
    https://en.wikipedia.org/wiki/National_identification_number#Philippines
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'PH',
        # length without insignificant chars
        'min_length': 12,
        'max_length': 12,
        'parsable': False,
        'checksum': False,
        'regexp': re.compile(r'^(\d{4}[ -]?\d{7}[ -]?\d)$'),
        'alias_of': None,
        'names': ['PhilID Card Number',
                  'PCN',
                  'PhilSys'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Philippines',
                  'https://en.wikipedia.org/wiki/Philippine_national_identity_card'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the PHL id number
        """
        return validate_regexp(id_number, PhilID.METADATA.regexp)



import re
from types import SimpleNamespace
from ..util import weighted_modulus_digit, validate_regexp


class NationalID:
    """
    SGP National ID number format, UIN, FIN
    https://en.wikipedia.org/wiki/National_identification_number#Singapore
    https://www.ngiam.net/NRIC/NRIC_numbers.pdf
    python version of https://github.com/IonBazan/NRIC
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'SG',
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^(?P<type>[STFGM])'
                             r'(?P<sn>\d{7})'
                             r'(?P<checksum>[A-Z])$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'NRIC',
                  'UIN',
                  'FIN'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Singapore',
                  'https://www.ngiam.net/NRIC/NRIC_numbers.pdf'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER = [2, 7, 6, 5, 4, 3, 2]
    CHECKSUM_MAP = {
        'S': 'JZIHGFEDCBA',
        'T': 'GFEDCBAJZIH',
        'F': 'XWUTRQPNMLK',
        'G': 'RQPNMLKXWUT',
        'M': 'XWUTRQPNJLK'
    }

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the SGP id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.checksum(id_number)

    @staticmethod
    def checksum(id_number) -> bool:
        """
        algorithm from:
        https://www.ngiam.net/NRIC/NRIC_numbers.pdf
        https://github.com/IonBazan/NRIC
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        series = id_number[0]
        checksum = id_number[-1]
        # it uses modulus 11 algorithm with magic numbers
        numbers = [int(char) for char in id_number[1:-1]]
        modulus = weighted_modulus_digit(numbers, NationalID.MAGIC_MULTIPLIER, 11, True)
        return checksum == NationalID.CHECKSUM_MAP[series][modulus]



import re
from math import floor
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..constant import Gender


def normalize(id_number: str) -> str:
    """make all characters to upper case"""
    return id_number.upper() if id_number else None


class ParseResult(TypedDict):
    """parse result for NationalID"""
    province_country_code: str
    """registration province/country code"""
    yyyy: int
    """year of birthday"""
    sn: str
    """serial number"""
    gender: Gender
    """gender, possible value: male, female"""


class NationalID:
    """
    Vietnam ID card
    https://vietnaminsider.vn/what-do-the-12-digits-on-the-citizen-id-card-with-chip-mean/
    https://lawnet.vn/en/vb/Circular-07-2016-TT-BCA-detailing-Law-on-Citizen-Identification-137-2015-ND-CP-5CCC3.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'VN',
        'min_length': 12,
        'max_length': 12,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<province_country_code>\d{3})'
                             r'(?P<gender>\d)'
                             r'(?P<yy>\d{2})'
                             r'(?P<sn>\d{6})$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'Thẻ căn cước công dân',
                  ],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Vietnam',
                  'https://vietnaminsider.vn/what-do-the-12-digits-on-the-citizen-id-card-with-chip-mean/',
                  'https://lawnet.vn/en/vb/Circular-07-2016-TT-BCA-detailing-Law-on-Citizen-Identification'
                  '-137-2015-ND-CP-5CCC3.html'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the VNM id number
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        province_country_code = match_obj.group('province_country_code')
        century_gender = match_obj.group('gender')
        yy = match_obj.group('yy')
        sn = match_obj.group('sn')
        yyyy = NationalID.get_birth_year(int(century_gender), int(yy))
        return {
            'province_country_code': province_country_code,
            'yyyy': yyyy,
            'sn': sn,
            'gender': Gender.MALE if int(century_gender) % 2 == 0 else Gender.FEMALE
        }

    @staticmethod
    def get_birth_year(century_gender: int, yy: int) -> int:
        return 1900 + 100 * floor(century_gender / 2) + yy



import re
from datetime import date, timedelta
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict
from ..constant import Gender
from ..util import weighted_modulus_digit, modulus_overflow_mod10, validate_regexp


class ParseResult(TypedDict):
    """parse result for the national id"""
    yyyymmdd: date
    """birthday"""
    gender: Gender
    """gender, possible value: male, female"""
    sn: str
    """serial number"""
    checksum: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """check digits"""


class NationalID:
    """
    LKA National ID number format
    # https://en.wikipedia.org/wiki/National_identification_number#Sri_Lanka
    # https://lk.linkedin.com/posts/nuwansenaratna_srilanka-activity-6926883712584335360-E_69
    # https://drp.gov.lk/Templates/Artical%20-%20English%20new%20number.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'LK',
        'min_length': 12,
        'max_length': 12,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<year>\d{4})'
                             r'(?P<days>\d{3})'
                             r'(?P<sn>\d{4})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['National ID Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Sri_Lanka',
                  'https://drp.gov.lk/Templates/Artical%20-%20English%20new%20number.html'],
        'deprecated': False
    })

    MAGIC_MULTIPLIER = [8, 4, 3, 2, 7, 6, 5, 7, 4, 3, 2]
    """multiplier for the checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the LKA id number
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        year = int(match_obj.group('year'))
        days = int(match_obj.group('days'))
        sn = match_obj.group('sn')
        checksum = NationalID.checksum(id_number)
        if not checksum:
            return None
        try:
            yyyymmdd = date(year, 1, 1) + timedelta(days - 501 if days > 500 else days - 1)
            return {
                'yyyymmdd': yyyymmdd,
                'gender': Gender.MALE if days < 500 else Gender.FEMALE,
                'sn': sn,
                'checksum': int(match_obj.group('checksum'))
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number) -> bool:
        """algorithm: https://lk.linkedin.com/posts/nuwansenaratna_srilanka-activity-6926883712584335360-E_69"""
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        # it uses modulus 11 algorithm with magic numbers
        numbers = [int(char) for char in id_number]
        modulus = modulus_overflow_mod10(weighted_modulus_digit(numbers[:-1], NationalID.MAGIC_MULTIPLIER, 11))
        return modulus == numbers[-1]



from .col.unique_persional_id import UniquePersonalID
from .util import alias_of

NationalID = alias_of(UniquePersonalID)
"""alias of UniquePersonalID"""
NUIP = alias_of(UniquePersonalID)
"""alias of UniquePersonalID"""



import re
from types import SimpleNamespace
from ..util import validate_regexp


class NationalInsuranceNumber:
    """
    UK National Insurance Number format
    https://en.wikipedia.org/wiki/National_Insurance_number
    https://www.gov.uk/hmrc-internal-manuals/national-insurance-manual/nim39110
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'GB',
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': False,
        'regexp': re.compile(r'^[A-Z]{2}\d{6}[A-Z]$'),
        'alias_of': None,
        'names': ['National Insurance Number',
                  'NI No',
                  'NINO'],
        'links': ['https://en.wikipedia.org/wiki/National_Insurance_number',
                  'https://www.gov.uk/hmrc-internal-manuals/national-insurance-manual/nim39110'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the GBR national insurance number
        """
        if not validate_regexp(id_number, NationalInsuranceNumber.METADATA.regexp):
            return False
        return (NationalInsuranceNumber.__check_prefix(id_number[:2]) and
                NationalInsuranceNumber.__check_suffix(id_number[-1]))

    @staticmethod
    def __check_prefix(prefix: str) -> bool:
        # These characters are not used as either the first or second letter of a NINO prefix
        prohibit_chars = ['D', 'F', 'I', 'Q', 'U', 'V']
        first_check = [True for i in prohibit_chars if i in prefix]
        if first_check:
            return False
        # The letter O is not used as the second letter of a prefix
        if 'O' in prefix[1]:
            return False
        # These codes are not to be used
        not_allocated_codes = ['BG', 'GB', 'NK', 'KN', 'TN', 'NT', 'ZZ']
        if prefix in not_allocated_codes:
            return False
        return True

    @staticmethod
    def __check_suffix(suffix: str) -> bool:
        allowed_suffix = ['A', 'B', 'C', 'D', 'F', 'M', 'P']
        return suffix in allowed_suffix



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import CHECK_DIGIT, weighted_modulus_digit, validate_regexp


class ParseResult(TypedDict):
    yyyymmdd: date
    """birthday of this ID"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """checksum"""


class CivilNumber:
    """
    Kuwait Civil Number, Arabic: الرقم المدني
    https://en.wikipedia.org/wiki/National_identification_number#Kuwait
    https://prakhar.me/articles/kuwait-civil-id-checksum/
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'KW',
        'min_length': 12,
        'max_length': 12,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<century>\d)'
                             r'(?P<yy>\d{2})'
                             r'(?P<mm>\d{2})'
                             r'(?P<dd>\d{2})'
                             r'(?P<sn>\d{4})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['Civil Number',
                  'الرقم المدني'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Kuwait',
                  'https://prakhar.me/articles/kuwait-civil-id-checksum/'],
        'deprecated': False
    })

    MULTIPLIER = [2, 1, 6, 3, 7, 9, 10, 5, 8, 4, 2]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the civil number
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return CivilNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """
        parse the id number
        """
        match_obj = CivilNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        checksum = CivilNumber.checksum(id_number)
        if checksum is None or str(checksum) != match_obj.group('checksum'):
            return None

        century = match_obj.group('century')
        if century == '2':
            year_base = 1900
        elif century == '3':
            year_base = 2000
        else:
            return None
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        return {
            'yyyymmdd': date(year_base + yy, mm, dd),
            'sn': match_obj.group('sn'),
            'checksum': checksum
        }

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        """
        https://prakhar.me/articles/kuwait-civil-id-checksum/
        """
        if not validate_regexp(id_number, CivilNumber.METADATA.regexp):
            return None

        numbers = [int(char) for char in id_number]
        modulus = weighted_modulus_digit(numbers[:-1], CivilNumber.MULTIPLIER, 11)
        if modulus > 10:
            # according to the algorithm, it will not be greater than 10
            return None
        return modulus



import re
from types import SimpleNamespace

from ..util import validate_regexp


class SocialSecurityNumber:
    """
    San Marino, individual social security number, SSI number
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/San-Marino-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'SM',
        'min_length': 9,
        'max_length': 9,
        # length without insignificant chars
        'parsable': False,
        # has parse function
        'checksum': False,
        # has checksum function
        'regexp': re.compile(r'^\d{9}$'),
        # regular expression to validate the id
        'alias_of': None,
        'names': ['Social Security Number',
                  'SSI'],
        'links': ['https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/San-Marino-TIN.pdf'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return validate_regexp(id_number, SocialSecurityNumber.METADATA.regexp)


class TaxRegistrationNumber:
    """
    San Marino, entity tax registration number, COE number
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/San-Marino-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'SM',
        'min_length': 7,
        'max_length': 7,
        # length without insignificant chars
        'parsable': False,
        # has parse function
        'checksum': False,
        # has checksum function
        'regexp': re.compile(r'^SM\d{5}$')
        # regular expression to validate the id
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return validate_regexp(id_number, TaxRegistrationNumber.METADATA.regexp)


SSI = SocialSecurityNumber
"""alias of SocialSecurityNumber"""



import re


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'-', '', id_number)


def is_new_personal_code(id_number: str) -> bool:
    """use the first 2 char to determine the new or old format"""
    try:
        return int(id_number[0:2]) > 31
    except ValueError:
        return False



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, Tuple, TypedDict
from ..constant import Gender
from ..util import CHECK_DIGIT, modulus_overflow_mod10, validate_regexp, weighted_modulus_digit


YEAR_MONTH_TYPE = Tuple[int, int]


class ParseResult(TypedDict):
    yyyymmdd: date
    """birthday of this ID"""
    gender: Gender
    """only male or female"""
    sn: str
    """serial number"""
    checksum: CHECK_DIGIT
    """checksum digit"""


class PESEL:
    """
    Poland PESEL code
    https://en.wikipedia.org/wiki/PESEL
    https://en.wikipedia.org/wiki/National_identification_number#Poland
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'PL',
        'min_length': 11,
        'max_length': 11,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yy>\d{2})'
                             r'(?P<mm>\d{2})'
                             r'(?P<dd>\d{2})'
                             r'(?P<sn>\d{4})'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['PESEL',
                  'Powszechny Elektroniczny System Ewidencji Ludności',
                  'Universal Electronic System for Registration of the Population'],
        'links': ['https://en.wikipedia.org/wiki/PESEL',
                  'https://en.wikipedia.org/wiki/National_identification_number#Poland'],
        'deprecated': False
    })

    MAGIC_NUMBERS = [1, 3, 7, 9, 1, 3, 7, 9, 1, 3]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the Poland PESEL code
        """
        if not id_number:
            return False

        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return PESEL.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """
        parse the id number
        """
        match_obj = PESEL.METADATA.regexp.match(id_number)
        if not match_obj:
            return None

        checksum = PESEL.checksum(id_number)
        if checksum is None or str(checksum) != match_obj.group('checksum'):
            return None

        yy = int(match_obj.group('yy'))
        mm_coded = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        year_base, mm = PESEL.get_year_base_month(mm_coded)
        sn = match_obj.group('sn')
        try:
            return {
                'yyyymmdd': date(year_base + yy, mm, dd),
                'gender': Gender.MALE if int(sn[-1]) % 2 == 1 else Gender.FEMALE,
                'sn': sn,
                'checksum': checksum
            }
        except ValueError:
            return None

    @staticmethod
    def get_year_base_month(month: int) -> YEAR_MONTH_TYPE:
        """ from https://en.wikipedia.org/wiki/PESEL#Birthdates """
        if month > 80:
            return 1800, month - 80
        elif month > 60:
            return 2200, month - 60
        elif month > 40:
            return 2100, month - 40
        elif month > 20:
            return 2000, month - 20
        else:
            return 1900, month

    @staticmethod
    def checksum(id_number) -> Optional[CHECK_DIGIT]:
        """
        python implementation of https://en.wikipedia.org/wiki/PESEL#Checksum_calculation
        """
        if not validate_regexp(id_number, PESEL.METADATA.regexp):
            return None
        numbers = [int(char) for char in id_number[:-1]]
        return modulus_overflow_mod10(weighted_modulus_digit(numbers, PESEL.MAGIC_NUMBERS, 10))



"""
The idnumbers package enables your application to parse, validate against to different specs. For example, it supports
[national ID](idnumbers/nationalid.html) for parsing and validating all citizen/resident ID numbers among different
countries.
"""



import re
from types import SimpleNamespace
from typing import Optional

from ..util import validate_regexp
from .old_national_id import OldNationalID, OldParseResult


class ParseResult(OldParseResult):
    yyyy: str
    """birth year"""


class NationalID(OldNationalID):
    """
    Bangladesh National ID number, জাতীয় পরিচয়পত্র, NID, BD
    https://en.wikipedia.org/wiki/National_identity_card_(Bangladesh)
    http://nationalidcardbangladesh.blogspot.com/2016/04/voter-id-national-id-card-number.html
    https://www.facebook.com/428195627559147/photos/a.428251897553520/428251617553548/?type=3
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'BD',
        # length without insignificant chars
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<yyyy>\d{4})'
                             r'(?P<distinct>\d{2})'
                             r'(?P<rmo>\d)'
                             r'(?P<police>\d{2})'
                             r'(?P<union>\d{2})'
                             r'(?P<sn>\d{6})$'),
        'alias_of': None,
        'names': ['Bangladesh national ID number',
                  'জাতীয় পরিচয়পত্র',
                  'NID',
                  'BD'],
        'links': [
            'https://en.wikipedia.org/wiki/National_identity_card_(Bangladesh)',
            'http://nationalidcardbangladesh.blogspot.com/2016/04/voter-id-national-id-card-number.html',
            'https://www.facebook.com/428195627559147/photos/a.428251897553520/428251617553548/?type=3'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """
        Parse the result
        """
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        old_result = OldNationalID.parse(id_number[4:])
        if not old_result:
            return None
        return {
            **old_result,
            'yyyy': int(id_number[0:4])
        }



from .fin.personal_id import PersonalIdentityCode
from .util import alias_of

NationalID = alias_of(PersonalIdentityCode)
"""alias of PersonalIdentityCode"""
HETU = alias_of(PersonalIdentityCode)
"""alias of PersonalIdentityCode"""



import re
from types import SimpleNamespace
from typing import Optional
from ..util import CHECK_DIGIT, alias_of, validate_regexp
from .util import normalize


class TaxFileNumber:
    """
    Australia tax file number format
    Note: Australian law specifically prohibits the use of the TFN as a national identification number
    https://en.wikipedia.org/wiki/National_identification_number#Australia
    https://en.wikipedia.org/wiki/Tax_file_number
    https://www.ato.gov.au/General/What-is-a-tax-file-number----Easy-Read/
    https://en-academic.com/dic.nsf/enwiki/436130
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'AU',
        # length without insignificant chars
        'min_length': 8,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(\d{9}|\d{8})$'),
        'alias_of': None,
        'names': ['Tax file number',
                  'TFN'],
        'links': [
            'https://en.wikipedia.org/wiki/National_identification_number#Australia',
            'https://en.wikipedia.org/wiki/Tax_file_number',
            'https://www.ato.gov.au/General/What-is-a-tax-file-number----Easy-Read/',
            'https://en-academic.com/dic.nsf/enwiki/436130'],
        'deprecated': False
    })

    MULTIPLIER = [1, 4, 3, 7, 5, 8, 6, 9, 10]
    """The magic multiplier for checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the AUS tax file number
        """
        if not validate_regexp(id_number, TaxFileNumber.METADATA.regexp):
            return False
        return TaxFileNumber.checksum(id_number) == 0

    @staticmethod
    def checksum(id_number: str) -> Optional[CHECK_DIGIT]:
        if not validate_regexp(id_number, TaxFileNumber.METADATA.regexp):
            return None
        """algorithm: https://en.wikipedia.org/wiki/Tax_file_number#Check_digit"""
        normalized = normalize(id_number)
        if len(normalized) == 8:
            normalized = normalized[0:7] + '0' + normalized[7]
        number_list = [int(char) for char in list(normalized)]
        return sum([value * TaxFileNumber.MULTIPLIER[index] for (index, value) in enumerate(number_list)]) % 11


TFN = alias_of(TaxFileNumber)
"""alias of TaxFileNumber"""



import re
from types import SimpleNamespace

from ..util import validate_regexp


class PersonalCode:
    """
    Moldova Personal Code, IDNP
    https://en.wikipedia.org/wiki/National_identification_number#Moldova
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'MD',
        # length without insignificant chars
        'min_length': 13,
        'max_length': 13,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': False,
        # regular expression to validate the id
        'regexp': re.compile(r'^\d{13}$'),
        'alias_of': None,
        'names': ['Personal Code',
                  'IDNP'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Moldova'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the personal code
        """
        return validate_regexp(id_number, PersonalCode.METADATA.regexp)



import re
from types import SimpleNamespace

from ..util import validate_regexp


class NationalID:
    """
    Nepal national identity card number, NIN
    https://en.wikipedia.org/wiki/National_identification_number#Nepal
    https://nimc.gov.ng/about-nin/
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'NP',
        'min_length': 11,
        'max_length': 11,
        # length without insignificant chars
        'parsable': False,
        # has parse function
        'checksum': False,
        # has checksum function
        'regexp': re.compile(r'^\d{11}$'),
        # regular expression to validate the id
        'alias_of': None,
        'names': ['National ID Number',
                  'NIN'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Nepal',
                  'https://nimc.gov.ng/about-nin/'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return validate_regexp(id_number, NationalID.METADATA.regexp)



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import validate_regexp


class ParseResult(TypedDict):
    """The parse result of Finland personal identity code, HETU"""
    yyyymmdd: date
    """Birthday"""
    sn: str
    """The serial number born at the same date"""


class PersonalIdentityNumber:
    """
    Denmark personal identity number, CPR, Det Centrale Personregister
    https://en.wikipedia.org/wiki/National_identification_number#Denmark
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'DK',
        'min_length': 10,
        'max_length': 10,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<dd>\d{2})(?P<mm>\d{2})(?P<yy>\d{2})-?'
                             r'(?P<sn>\d{4})$'),
        'alias_of': None,
        'names': ['personal identity number',
                  'CPR',
                  'Det Centrale Personregister'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Denmark'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the id number
        """
        if not validate_regexp(id_number, PersonalIdentityNumber.METADATA.regexp):
            return False
        return PersonalIdentityNumber.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """ parse the CPR id"""
        match_obj = PersonalIdentityNumber.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        sn = match_obj.group('sn')
        yyyy_base = 1900 if yy > 50 else 2000
        try:
            return {
                'yyyymmdd': date(yyyy_base + yy, mm, dd),
                'sn': sn
            }
        except ValueError:
            return None



import re
from types import SimpleNamespace
from ..util import validate_regexp, weighted_modulus_digit


class CivilIDNumber:
    """
    Portugal Civil ID number
    Número de identificação civil or NIC or BI
    https://en.wikipedia.org/wiki/National_identification_number#Portugal
    https://www.atractor.pt/mat/alg_controlo/bifm2-_en.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'PT',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^(\d{9})$'),
        'alias_of': None,
        'names': ['Civil ID Number',
                  'Número de identificação civil',
                  'NIC',
                  'BI'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Portugal',
                  'https://www.atractor.pt/mat/alg_controlo/bifm2-_en.html'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the PRT civil id number
        """
        if not validate_regexp(id_number, CivilIDNumber.METADATA.regexp):
            return False
        return CivilIDNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        multipliers = [9, 8, 7, 6, 5, 4, 3, 2]
        mod = weighted_modulus_digit([int(i) for i in id_number[:-1]], multipliers, 11, True)
        calculated_checksum = 0 if mod == 0 or mod == 1 else (11 - mod)
        return calculated_checksum == int(id_number[-1])



from typing import Union
from .lva.personal_code import PersonalCode
from .lva.old_personal_code import OldPersonalCode
from .util import alias_of


def get_validator(id_number: str) -> Union[type[PersonalCode], type[OldPersonalCode]]:
    """check the first 2 char to return the correct class"""

    try:
        return PersonalCode if int(id_number[0:2]) > 31 else OldPersonalCode
    except ValueError:
        return OldPersonalCode


NationalID = alias_of(PersonalCode)
"""alias of PersonalCode"""



import re
from types import SimpleNamespace
from ..util import validate_regexp


class NationalID:
    """
    Iraq National Card number. not enough docs to research.
    https://en.wikipedia.org/wiki/Iraq_National_Card
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'IQ',
        'min_length': 12,
        'max_length': 12,
        'parsable': False,
        'checksum': False,
        'regexp': re.compile(r'^\d{12}$'),
        'alias_of': None,
        'names': ['National Card Number',
                  'البطاقة الوطنية',
                  'كارتى نيشتمانى'],
        'links': ['https://en.wikipedia.org/wiki/Iraq_National_Card'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return validate_regexp(id_number, NationalID.METADATA.regexp)



import re
from datetime import date
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import CHECK_DIGIT, validate_regexp, luhn_digit, verhoeff_check


class ParseResult(TypedDict):
    """The parse result of Luxembourg ID"""
    yyyymmdd: date
    """dob"""
    sn: str
    """serial number"""
    checksum1: CHECK_DIGIT
    """check digit"""
    checksum2: CHECK_DIGIT
    """check digit"""


class NationalID:
    """
    Luxembourg National ID number format
    https://en.wikipedia.org/wiki/National_identification_number#Luxembourg
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Luxembourg-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'LU',
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<yyyy>\d{4})(?P<mm>\d{2})(?P<dd>\d{2})'
                             r'(?P<sn>\d{3})'
                             r'(?P<checksum1>\d)'
                             r'(?P<checksum2>\d)$'),
        'alias_of': None,
        'names': ['National ID Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Luxembourg',
                  'https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/Luxembourg-TIN.pdf'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate"""
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = NationalID.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        checksum = NationalID.checksum(id_number)
        if not checksum:
            return None
        yyyy = int(match_obj.group('yyyy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        try:
            return {
                'yyyymmdd': date(yyyy, mm, dd),
                'sn': match_obj.group('sn'),
                'checksum1': int(match_obj.group('checksum1')),
                'checksum2': int(match_obj.group('checksum2'))
            }
        except ValueError:
            # catch the date error
            return None

    @staticmethod
    def checksum(id_number) -> bool:
        """check the checksum"""
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        numbers = [int(char) for char in id_number]
        check1 = luhn_digit(numbers[0:-2], True)
        if check1 != numbers[-2]:
            return False
        # clone the numbers and remove the 12th digit
        check2_numbers = list(numbers)
        del check2_numbers[-2]
        # perform verhoeff check
        return verhoeff_check(check2_numbers)



import re
from types import SimpleNamespace

from ..util import validate_regexp


class IDCardNumber:
    """
    Venezuela ID card number
    https://en.wikipedia.org/wiki/National_identification_number#Venezuela
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'VE',
        # length without insignificant chars
        'min_length': 9,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': False,
        # regular expression to validate the id
        'regexp': re.compile(r'^V ?\d{2}\.?\d{3}\.?\d{3}$'),
        'alias_of': None,
        'names': ['ID Card Number',
                  'Cédula de Identidad'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Venezuela'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return validate_regexp(id_number, IDCardNumber.METADATA.regexp)



import re
from types import SimpleNamespace
from ..util import validate_regexp, weighted_modulus_digit


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r'[-.]', '', id_number)


class NationalID:
    """
    CHL national ID number format, RUN (Rol Único Nacional), RUT (Rol Único Tributario)
    https://en.wikipedia.org/wiki/National_identification_number#Chile
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'CL',
        # length without insignificant chars
        'min_length': 8,
        'max_length': 9,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(\d{1,2}\.\d{3}\.\d{3}-[\d|K])$'),
        'alias_of': None,
        'names': ['Rol Único Nacional',
                  'RUN',
                  'Rol Único Tributario',
                  'RUT'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Chile'],
        'deprecated': False
    })

    MULTIPLIER = [3, 2, 7, 6, 5, 4, 3, 2]

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the CHL id number
        https://codepen.io/alisteroz/pen/KEoqgQ
        """
        if not validate_regexp(id_number, NationalID.METADATA.regexp):
            return False
        return NationalID.checksum(id_number) == id_number[-1]

    @staticmethod
    def checksum(id_number: str) -> str:
        """
        Validate CHL national id number checksum
        https://gist.github.com/ryangreenberg/4531891
        """
        number_list = [int(char) for char in list(normalize(id_number)[:-1])]
        modulus = weighted_modulus_digit(number_list, NationalID.MULTIPLIER, 11)
        return str(0 if modulus == 11 else 'K' if modulus == 10 else modulus)



import re
from datetime import date
from types import SimpleNamespace
from typing import Literal, Optional, TypedDict
from ..util import validate_regexp
from ..constant import Citizenship, Gender


class ParseResult(TypedDict):
    """parse result for national id"""
    yyyymmdd: date
    """birthday"""
    gender: Gender
    """gender, possible value: male, female"""
    citizenship: Citizenship
    """citizenship"""
    sn: str
    """serial number if two persons born on the same date"""


class OldIDParseResult(ParseResult):
    """Old format contains more information"""
    location: str
    """registration location"""
    checksum: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """check digits"""


class ResidentRegistration:
    """
    KOR resident registration number format. The ARC is the same as ResidentRegistration. KOR removed the checksum and location
    from Oct. 2020 to protect privacy.
    # https://en.wikipedia.org/wiki/Resident_registration_number
    # https://centers.ibs.re.kr/html/living_en/overview/arc.html
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'KR',
        'min_length': 13,
        'max_length': 13,
        'parsable': True,
        'checksum': False,
        'regexp': re.compile(r'^(?P<yy>\d{2})(?P<mm>\d{2})(?P<dd>\d{2})-'
                             r'(?P<gender>\d)'
                             r'(?P<sn>\d{6})$'),
        'alias_of': None,
        'names': ['Resident Registration Number',
                  '주민등록번호',
                  'RRN',
                  '住民登錄番號',
                  'Jumin Deungnok Beonho'],
        'links': ['https://en.wikipedia.org/wiki/Resident_registration_number',
                  'https://centers.ibs.re.kr/html/living_en/overview/arc.html'],
        'deprecated': False
    })

    CITIZENSHIP_MAP = {
        9: Citizenship.CITIZEN,
        0: Citizenship.CITIZEN,
        1: Citizenship.CITIZEN,
        2: Citizenship.CITIZEN,
        3: Citizenship.CITIZEN,
        4: Citizenship.CITIZEN,
        5: Citizenship.RESIDENT,
        6: Citizenship.RESIDENT,
        7: Citizenship.RESIDENT,
        8: Citizenship.RESIDENT
    }
    """citizenship map"""

    DOB_BASE_MAP = {
        9: 1800,
        0: 1800,
        1: 1900,
        2: 1900,
        3: 2000,
        4: 2000,
        5: 1900,
        6: 1900,
        7: 2000,
        8: 2000
    }
    """dob base year map"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the KOR id number
        """
        if not validate_regexp(id_number, ResidentRegistration.METADATA.regexp):
            return False
        return ResidentRegistration.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the result"""
        match_obj = ResidentRegistration.METADATA.regexp.match(id_number)
        return ResidentRegistration.build_parse_result(match_obj)

    @staticmethod
    def build_parse_result(match_obj: re.Match[str]) -> Optional[ParseResult]:
        if not match_obj:
            return None
        yy = int(match_obj.group('yy'))
        mm = int(match_obj.group('mm'))
        dd = int(match_obj.group('dd'))
        gender = int(match_obj.group('gender'))
        sn = match_obj.group('sn')
        yyyy_base = ResidentRegistration.DOB_BASE_MAP[gender]
        try:
            return {
                'yyyymmdd': date(yyyy_base + yy, mm, dd),
                'gender': Gender.MALE if gender % 2 == 1 else Gender.FEMALE,
                'citizenship': ResidentRegistration.CITIZENSHIP_MAP[gender],
                'sn': sn
            }
        except ValueError:
            return None



from .alb.identity_number import IdentityNumber
from .util import alias_of

NationalID = alias_of(IdentityNumber)
"""alias of IdentityNumber"""
NISH = alias_of(IdentityNumber)
"""alias of IdentityNumber"""
NIPT = alias_of(IdentityNumber)
"""alias of IdentityNumber"""
NID = alias_of(IdentityNumber)
"""alias of IdentityNumber"""



import re
from types import SimpleNamespace
from ..util import validate_regexp, ean13_digit


def normalize(id_number: str) -> str:
    """strip out useless characters/whitespaces"""
    return re.sub(r'\.', '', id_number)


class SocialSecurityNumber:
    """
    Switzerland Social Security Number (AHV-Nr. [de] / No AVS [fr])
    https://en.wikipedia.org/wiki/National_identification_number#Switzerland
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'CH',
        # length without insignificant chars
        'min_length': 13,
        'max_length': 13,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^756\.\d{4}\.\d{4}.\d{2}$'),
        'alias_of': None,
        'names': ['Social Security Number',
                  'AHV-Nr.',
                  'No AVS'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Switzerland'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """validate the number"""
        return SocialSecurityNumber.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """use EAN-13 to validate the number"""
        if not validate_regexp(id_number, SocialSecurityNumber.METADATA.regexp):
            return False
        numbers = [int(char) for char in normalize(id_number)]
        return numbers[-1] == ean13_digit(numbers[:-1])



import re
from types import SimpleNamespace
from typing import Optional, TypedDict
from ..util import validate_regexp
from ..constant import Gender


class BirthDepartment(TypedDict):
    """the commune of origin"""
    department: str
    city: str
    country: str


class ParseResult(TypedDict):
    """parse result of INSEE"""
    gender: Gender
    """gender, possible value: male, female"""
    yy: str
    """year of birth"""
    mm: str
    """month of birth"""
    birth_department: BirthDepartment
    """the commune of origin"""
    checksum: str
    """checksum string"""


class INSEE:
    """
    France National ID number, INSEE
    https://en.wikipedia.org/wiki/National_identification_number#France
    https://fr.wikipedia.org/wiki/Num%C3%A9ro_de_s%C3%A9curit%C3%A9_sociale_en_France#Signification_des_chiffres_du_NIR
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'FR',
        # length without insignificant chars
        'min_length': 15,
        'max_length': 15,
        'parsable': True,
        'checksum': True,
        'regexp': re.compile(r'^(?P<gender>([123478]))'
                             r'(?P<yy>\d{2})'
                             r'(?P<mm>(0[1-9]|1[0-2]|[2-3][0-9]|4[0-2]|[5-9][0-9]))'
                             r'(?P<birth_department>((\d{2}|2[AaBb])\d{3}))'
                             r'(?P<cert_number>((?!000)\d{3}))'
                             r'(?P<control_key>((?!(00|98|99))\d{2}))$'),
        'alias_of': None,
        'names': ['National ID Number',
                  'INSEE',
                  'NIR',
                  'NIRPP'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#France',
                  'https://fr.wikipedia.org/wiki/'
                  'Num%C3%A9ro_de_s%C3%A9curit%C3%A9_sociale_en_France#Signification_des_chiffres_du_NIR'],
        'deprecated': False

    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the FRA id number
        """
        if not validate_regexp(id_number, INSEE.METADATA.regexp):
            return False
        if not INSEE.parse(id_number):
            return False
        return INSEE.checksum(id_number)

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        match_obj = INSEE.METADATA.regexp.match(id_number)
        if not match_obj:
            return None
        birth_department = INSEE.validate_birth_department(match_obj.group('birth_department'))
        if not birth_department:
            return None
        gender = match_obj.group('gender')
        yy = match_obj.group('yy')
        mm = match_obj.group('mm')
        control_key = match_obj.group('control_key')
        return {
            'gender': Gender.MALE if gender == '1' else Gender.FEMALE,
            'yy': yy,
            'mm': mm,
            'birth_department': birth_department,
            'checksum': control_key,
        }

    @staticmethod
    def checksum(id_number: str) -> bool:
        """algorithm: https://en.wikipedia.org/wiki/INSEE_code#National_identification_numbers"""
        normalized = id_number.upper().replace('2A', '19').replace('2B', '18')
        return 97 - int(normalized[:-2]) % 97 == int(normalized[-2:])

    @staticmethod
    def validate_birth_department(birth_department: str) -> Optional[BirthDepartment]:
        department_code = birth_department[:2].upper()
        if (department_code.isdigit() and 1 <= int(department_code) <= 95) or department_code in ['2A', '2B']:
            return {
                "department": birth_department[:2],
                "city": birth_department[2:],
                "country": ""
            }
        if 97 <= int(department_code) <= 98:
            return {
                "department": birth_department[:3],
                "city": birth_department[3:],
                "country": ""
            }
        if department_code == '99':
            return {
                "department": '',
                "city": '',
                "country": birth_department[2:]
            }
        return None



import re
from types import SimpleNamespace

from ..util import validate_regexp


class CitizenIDNumber:
    """
    Slovakia Citizen Identification Card Number format, Číslo občianskeho preukazu (ČOP)
    https://en.wikipedia.org/wiki/National_identification_number#Slovakia
    https://en.wikipedia.org/wiki/Slovak_identity_card
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'SK',
        'min_length': 8,
        'max_length': 8,
        # length without insignificant chars
        'parsable': False,
        # has parse function
        'checksum': False,
        # has checksum function
        'regexp': re.compile(r'^[A-Z]{2} ?\d{6}$'),
        # regular expression to validate the id
        'alias_of': None,
        'names': ['Citizen Identification Card Number',
                  'Číslo občianskeho preukazu',
                  'ČOP'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Slovakia',
                  'https://en.wikipedia.org/wiki/Slovak_identity_card'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate CitizenIDNumber
        """
        return validate_regexp(id_number, CitizenIDNumber.METADATA.regexp)



import re
from types import SimpleNamespace
from typing import List
from ..util import CHECK_DIGIT, mn_modulus_digit, modulus_overflow_mod10, validate_regexp


def normalize(id_number):
    """strip out useless characters/whitespaces"""
    return re.sub(r' ', '', id_number)


class TaxID:
    """
    Germany Tax ID, Steuerliche Identifikationsnummer, Persönliche Identificationsnummer, Identifikationsnummer,
    Steuer-IdNr., IdNr or Steuer-ID.
    https://allaboutberlin.com/guides/german-tax-id-steuernummer
    python version of https://github.com/kontist/validate-steuerid
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'DE',
        'min_length': 11,
        'max_length': 11,
        'parsable': False,
        'checksum': True,
        'regexp': re.compile(r'^\d{2} ?\d{3} ?\d{3} ?\d{3}$'),
        'alias_of': None,
        'names': ['Tax ID',
                  'Steuerliche Identifikationsnummer',
                  'Persönliche Identificationsnummer',
                  'Identifikationsnummer',
                  'Steuer-IdNr.',
                  'IdNr',
                  'Steuer-ID'],
        'links': ['https://allaboutberlin.com/guides/german-tax-id-steuernummer'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the DEU Tax ID
        """
        if not validate_regexp(id_number, TaxID.METADATA.regexp):
            return False
        elif not TaxID.check_multiple_occurrence(id_number):
            return False
        elif not TaxID.check_consecutive_position(id_number):
            return False
        return TaxID.checksum(id_number)

    @staticmethod
    def checksum(id_number: str) -> bool:
        """check if the ID valid against its checksum"""
        if not validate_regexp(id_number, TaxID.METADATA.regexp):
            return False
        numbers = [int(char) for char in normalize(id_number)]
        check = numbers[-1]
        return int(check) == TaxID.get_checkdigit(numbers[:-1])

    @staticmethod
    def get_checkdigit(numbers: List[int]) -> CHECK_DIGIT:
        return modulus_overflow_mod10(mn_modulus_digit(numbers, 10, 11))

    @staticmethod
    def check_multiple_occurrence(id_number: str) -> bool:
        """
        We can only have 1 digit with multiple occurrence in the number
        """
        normalized = list(normalize(id_number))[:-1]
        normalized.sort()
        first_multiple_digits = None
        for index, digit in enumerate(normalized):
            if index == 0:
                continue
            if not first_multiple_digits and digit == normalized[index - 1]:
                # no first occurrence, we set the digit to the first_multiple_digits
                first_multiple_digits = digit
            elif first_multiple_digits != digit and digit == normalized[index - 1]:
                # we already saw first_multiple_digits and find it again. it's wrong
                return False
            # first_multiple_digits == digit and digit == normalized[index - 1] means the 3rd occurrence. It's ok.
        return True

    @staticmethod
    def check_consecutive_position(id_number: str) -> bool:
        """
        We can have a number with 2 consecutive position at most.
        """
        normalized = list(normalize(id_number))
        # minus one for getting rid of check digit
        for index in range(len(normalized) - 2):
            if normalized[index] == normalized[index + 1] and normalized[index] == normalized[index + 2]:
                return False
        return True



from .dnk.personal_id import PersonalIdentityNumber
from .util import alias_of

NationalID = alias_of(PersonalIdentityNumber)
"""alias of PersonalIdentityNumber"""
CPR = alias_of(PersonalIdentityNumber)
"""alias of PersonalIdentityNumber"""



import re
from types import SimpleNamespace
from typing import Optional

from ..util import CHECK_DIGIT, validate_regexp
from .util import normalize


class PersonalCode:
    """
    Latvia Personal Code format, personas kods
    https://en.wikipedia.org/wiki/National_identification_number#Latvia
    https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Latvia-TIN.pdf
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'LV',
        # length without insignificant chars
        'min_length': 11,
        'max_length': 11,
        # has parse function
        'parsable': False,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(\d{6}-?\d{5}$)'),
        'alias_of': None,
        'names': ['Personal Code',
                  'personas kods'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#Latvia',
                  'https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/'
                  'tax-identification-numbers/Latvia-TIN.pdf'],
        'deprecated': False
    })

    MULTIPLIER = [1, 6, 3, 7, 9, 10, 5, 8, 4, 2]
    """multiplier for checksum"""

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate
        """
        return str(PersonalCode.checksum(id_number)) == id_number[-1]

    @staticmethod
    def checksum(id_number: str) -> Optional[CHECK_DIGIT]:
        if not validate_regexp(id_number, PersonalCode.METADATA.regexp):
            return None
        """Calculate national id checksum: (1101-sum) mod 11 and mod 10"""
        numbers = [int(i) for i in normalize(id_number)[:10]]
        weighted_value = sum([value * PersonalCode.MULTIPLIER[index] for (index, value) in enumerate(numbers)])
        return (1101 - weighted_value) % 11 % 10



import re
from datetime import date
from typing import Optional, TypedDict
from types import SimpleNamespace
from ..constant import Citizenship, Gender
from ..util import CHECK_DIGIT, luhn_digit


class ParseResult(TypedDict):
    """parse result of national id"""
    yyyymmdd: date
    """birthday"""
    sn: str
    """serial number"""
    gender: Gender
    """gender, possible value: male, female"""
    citizenship: Citizenship
    """citizenship"""
    checksum: CHECK_DIGIT
    """check digit"""


class NationalID:
    """
    South Africa national ID number format
    https://en.wikipedia.org/wiki/National_identification_number#South_Africa
    https://www.westerncape.gov.za/general-publication/decoding-your-south-african-id-number-0
    """
    METADATA = SimpleNamespace(**{
        'iso3166_alpha2': 'ZA',
        # length without insignificant chars
        'min_length': 13,
        'max_length': 13,
        # has parse function
        'parsable': True,
        # has checksum function
        'checksum': True,
        # regular expression to validate the id
        'regexp': re.compile(r'^(?P<yy>\d{2})(?P<mm>0[1-9]|1[12])'
                             r'(?P<dd>0[1-9]|[12][0-9]|3[01])'
                             r'(?P<sn>\d{4})'
                             r'(?P<citizenship>[01])([89])'
                             r'(?P<checksum>\d)$'),
        'alias_of': None,
        'names': ['National ID Number'],
        'links': ['https://en.wikipedia.org/wiki/National_identification_number#South_Africa',
                  'https://www.westerncape.gov.za/general-publication/decoding-your-south-african-id-number-0'],
        'deprecated': False
    })

    @staticmethod
    def validate(id_number: str) -> bool:
        """
        Validate the ZAF id number
        """
        if not isinstance(id_number, str):
            id_number = repr(id_number)
        return NationalID.parse(id_number) is not None

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        match_obj = NationalID.METADATA.regexp.match(id_number)
        check_digit = NationalID.checksum(id_number)
        if not match_obj:
            return None
        elif check_digit != int(id_number[-1:]):
            return None

        year = int(match_obj.group('yy'))
        year += 2000 if year < 50 else 1900
        try:
            return {
                'yyyymmdd': date(year, int(match_obj.group('mm')), int(match_obj.group('dd'))),
                'sn': match_obj.group('sn'),
                'gender': Gender.MALE if int(match_obj.group('sn')[0]) > 4 else Gender.FEMALE,
                'citizenship': Citizenship.CITIZEN if match_obj.group('citizenship') == '0' else Citizenship.RESIDENT,
                'checksum': check_digit
            }
        except ValueError:
            return None

    @staticmethod
    def checksum(id_number: str) -> CHECK_DIGIT:
        """
        use Luhn algorithm.
        """
        return luhn_digit([int(char) for char in id_number[:-1]])



from copy import copy
from typing import Optional, Tuple
from ..constant import Citizenship
from ..yugoslavia import ParseResult, UniqueMasterCitizenNumber as YugoslaviaJMBG
from ..util import alias_of

BIH_METADATA = copy(YugoslaviaJMBG.METADATA)
BIH_METADATA.iso3166_alpha2 = 'BA'


class UniqueMasterCitizenNumber(YugoslaviaJMBG):
    """
    Serbia Unique Master Citizen Number format, JMBG
    https://en.wikipedia.org/wiki/Unique_Master_Citizen_Number
    """
    METADATA = BIH_METADATA

    @staticmethod
    def parse(id_number: str) -> Optional[ParseResult]:
        """parse the value"""
        result = YugoslaviaJMBG.parse(id_number)
        if not result:
            return None
        loc_citizenship = UniqueMasterCitizenNumber.check_location(result['location'])
        if not loc_citizenship:
            return None
        citizenship, location = loc_citizenship
        result['citizenship'] = citizenship
        return result

    @staticmethod
    def check_location(location: str) -> Optional[Tuple[Citizenship, str]]:
        result = YugoslaviaJMBG.check_location(location)
        if not result:
            return None
        """
        Since the North Macedonia is an independent country, they share the same id code base. So, the citizenship
        is 10 <= location < 20
        """
        if 10 <= int(location) < 20:
            return Citizenship.CITIZEN, location
        return Citizenship.RESIDENT, location


JMBG = alias_of(UniqueMasterCitizenNumber)
"""
Alias of UniqueMasterCitizenNumber
"""
